/*
 * main.cxx
 *
 * Dump PDU application to test and debug the ASN PER encoding of PDU's
 * in the Open H323 project.
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 * $Log: main.cxx,v $
 * Revision 1.5  1999/08/31 13:30:20  robertj
 * Added gatekeeper support.
 *
 * Revision 1.4  1999/08/14 03:26:45  robertj
 * Compiler compatibility problems.
 *
 * Revision 1.3  1999/08/09 14:06:32  robertj
 * Fixed hex reading for non-tcpdump file
 *
 * Revision 1.2  1999/08/09 04:50:57  robertj
 * Allowed for multiple input files and fixed program name in usage.
 *
 * Revision 1.1  1999/08/08 10:05:40  robertj
 * Added PDU dump program.
 *
 */

#include <ptlib.h>
#include <ptlib/sockets.h>
#include "main.h"


///////////////////////////////////////////////////////////////////////////////

TcpStreamKey::TcpStreamKey()
{
  tx_port = rx_port = 0;
  pdu_count = 0;
}


TcpStreamKey::TcpStreamKey(PIPSocket::Address _tx_ip,
                           unsigned           _tx_port,
                           PIPSocket::Address _rx_ip,
                           unsigned           _rx_port)
  : tx_ip(_tx_ip), tx_port(_tx_port),
    rx_ip(_rx_ip), rx_port(_rx_port)
{
  pdu_count = 0;
}


PObject::Comparison TcpStreamKey::Compare(const PObject & obj) const
{
  const TcpStreamKey & other = (const TcpStreamKey &)obj;
  if (tx_ip < other.tx_ip)
    return LessThan;
  if (tx_ip > other.tx_ip)
    return GreaterThan;
  if (tx_port < other.tx_port)
    return LessThan;
  if (tx_port > other.tx_port)
    return GreaterThan;
  if (rx_ip < other.rx_ip)
    return LessThan;
  if (rx_ip > other.rx_ip)
    return GreaterThan;
  if (rx_port < other.rx_port)
    return LessThan;
  if (rx_port > other.rx_port)
    return GreaterThan;
  return EqualTo;
}


PINDEX TcpStreamKey::HashFunction() const
{
  return ((DWORD)tx_ip+tx_port+(DWORD)rx_ip+rx_port)%23;
}


PObject * TcpStreamKey::Clone() const
{
  return new TcpStreamKey(tx_ip, tx_port, rx_ip, rx_port);
}


void TcpStreamKey::ReadFrom(istream & strm)
{
  char dummy;
  strm >> tx_ip >> dummy >> tx_port >> ws >> dummy
       >> rx_ip >> dummy >> rx_port >> dummy;
}


void TcpStreamKey::ReverseCopy(const TcpStreamKey & key)
{
  rx_ip   = key.tx_ip;
  rx_port = key.tx_port;
  tx_ip   = key.rx_ip;
  tx_port = key.rx_port;
  pdu_count = 0;
}


///////////////////////////////////////////////////////////////////////////////

TcpStreamData::TcpStreamData()
{
  position = 0;
}


void TcpStreamData::ReadFrom(istream & strm)
{
  SetSize(100);

  char hex[3];
  hex[2] = '\0';
  PINDEX count = 0;
  while (strm.good()) {
    strm >> ws >> hex[0] >> hex[1];
    theArray[count++] = (BYTE)strtoul(hex, NULL, 16);
    if (count >= GetSize())
      SetSize(count+100);
  }

  SetSize(count);
}


BOOL TcpStreamData::HasTPKT()
{
  PINDEX left = GetSize() - position;
  if (left <= 4)
    return FALSE;

  if (theArray[position] != 3)
    return FALSE;

  PINDEX tpkt_length = ((BYTE)theArray[position+2]<<8) + (BYTE)theArray[position+3];
  return left >= tpkt_length;
}


PBYTEArray TcpStreamData::GetTPKT()
{
  return PBYTEArray((BYTE *)theArray + position + 4,
                    ((BYTE)theArray[position+2]<<8) + (BYTE)theArray[position+3] - 4);
}


///////////////////////////////////////////////////////////////////////////////

void H225withH245::PrintOn(ostream & output) const
{
  int indent = output.precision();
  H225_H323_UserInformation::PrintOn(output);

  for (PINDEX i = 0; i < m_h323_uu_pdu.m_h245Control.GetSize(); i++) {
    output << '\n' << setw(indent) << " " << "H.245 PDU:";

    PPER_Stream per = m_h323_uu_pdu.m_h245Control[i].GetValue();
    H245_MultimediaSystemControlMessage h245;
    if (!h245.Decode(per))
      output << " Tunnelled H.245 decode failure." << endl;

    output << '\n' << setw(indent+2) << " " << setprecision(indent+2) << h245 << endl;
  }
}


///////////////////////////////////////////////////////////////////////////////

BOOL Q931withH225::Decode(const PBYTEArray & rawData)
{
  if (!Q931::Decode(rawData))
    return FALSE;

  if (HasField(Q931::UserUserField)) {
    PPER_Stream per = GetField(Q931::UserUserField);
    hasH225 = h225.Decode(per);
  }
  else
    hasH225 = FALSE;

  return TRUE;
}


void Q931withH225::PrintOn(ostream & output) const
{
  int indent = output.precision();
  Q931::PrintOn(output);

  if (hasH225)
    output << '\n'
           << setw(indent) << " " << "H.225 PDU:\n"
           << setw(indent+2) << " " << setprecision(indent+2) << h225;
}


///////////////////////////////////////////////////////////////////////////////

BOOL ReadPacket(istream & input, TcpStreamKey & key, PBYTEArray & payload, BOOL & isUDP)
{
  input.clear();

  // Skip over any packet hex dump lines (lines stating with whitespace)
  while (isspace(input.peek())) {
    input.ignore(1000, '\n');
    if (!input)
      return FALSE;
  }

  // Read the data from the tcpdump line
  char type_flag;
  PTimeInterval timestamp;
  input >> dec >> timestamp >> key >> type_flag;
  if (!input) {
    input.clear(input.rdstate()&~ios::failbit);
    input.ignore(1000, '\n');
    return FALSE;
  }

  PINDEX payload_length;

  switch (type_flag) {
    case '.' : // ack only packet, ignore
      input.ignore(1000, '\n');
      return FALSE;

    case 'u' : // udp packet
      input.ignore(2);
      input >> payload_length;
      isUDP = TRUE;
      break;

    default : // Other TCP packets
      // Look for length in parentheses
      input.ignore(50, '(');
      input >> payload_length;
      isUDP = FALSE;
  }

  if (!input)
    return FALSE;

  payload.SetSize(payload_length);

  // Waste rest of line
  input.ignore(1000, '\n');
  if (!input)
    return FALSE;

  // Finished here if no payload
  if (payload_length == 0)
    return TRUE;

  // Now read the hex dump part
  unsigned temp;
  input >> hex >> temp;
  if (!input)
    return FALSE;

  // Skip over the IP and TCP/UDP header
  unsigned header_length = ((temp >> 8)&15)*2 + (isUDP ? 4 : 10);
  while (--header_length > 0) {
    input >> temp;
    if (!input)
      return FALSE;
  }

  // Start reading the payload
  char hex[3];
  hex[2] = '\0';
  PINDEX count;
  for (count = 0; count < payload_length; count++) {
    input >> ws >> hex[0] >> hex[1];
    if (!input)
      return FALSE;
    payload[count] = (BYTE)strtoul(hex, NULL, 16);
  }

  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////

void DumpAllQ931(ostream & output,
                 TcpStreamDict & tcpStream,
                 const TcpStreamKey & key,
                 BOOL callerToReceiver,
                 TcpStreamKey & receiverControlChannel)
{
  if (!tcpStream.Contains(key))
    return;

  if (!tcpStream[key].HasTPKT())
    return;

  const char * direction = callerToReceiver ? " (caller -> receiver) "
                                            : " (receiver -> caller) ";

  Q931withH225 q931;
  if (!q931.Decode(tcpStream[key].GetTPKT())) {
    output << "Read Q.931" << direction << "failed.\n";
    return;
  }

  output << "Q.931 " << direction << " PDU:\n  " << setprecision(2) << q931 << endl;

  // Look for the opening of the H.245 channel
  if (!q931.hasH225)
    return;

  H225_TransportAddress * taddr = NULL;

  switch (q931.h225.m_h323_uu_pdu.m_h323_message_body.GetTag()) {
    case H225_H323_UU_PDU_h323_message_body::e_callProceeding :
    {
      H225_CallProceeding_UUIE & uu = q931.h225.m_h323_uu_pdu.m_h323_message_body;
      if (uu.HasOptionalField(H225_CallProceeding_UUIE::e_h245Address))
        taddr = &uu.m_h245Address;
      break;
    }

    case H225_H323_UU_PDU_h323_message_body::e_connect :
    {
      H225_Connect_UUIE & uu = q931.h225.m_h323_uu_pdu.m_h323_message_body;
      if (uu.HasOptionalField(H225_Connect_UUIE::e_h245Address))
        taddr = &uu.m_h245Address;
      break;
    }

    case H225_H323_UU_PDU_h323_message_body::e_alerting :
    {
      H225_Alerting_UUIE & uu = q931.h225.m_h323_uu_pdu.m_h323_message_body;
      if (uu.HasOptionalField(H225_Alerting_UUIE::e_h245Address))
        taddr = &uu.m_h245Address;
      break;
    }

    case H225_H323_UU_PDU_h323_message_body::e_facility :
    {
      H225_Facility_UUIE & uu = q931.h225.m_h323_uu_pdu.m_h323_message_body;
      if (uu.HasOptionalField(H225_Facility_UUIE::e_h245Address))
        taddr = &uu.m_h245Address;
      break;
    }
  }

  if (taddr == NULL)
    return;

  if (taddr->GetTag() != H225_TransportAddress::e_ipAddress)
    return;

  H225_TransportAddress_ipAddress & ipaddr = *taddr;
  memcpy(&receiverControlChannel.tx_ip, ipaddr.m_ip.GetPointer(), 4);
  receiverControlChannel.tx_port = ipaddr.m_port;
}


///////////////////////////////////////////////////////////////////////////////

void DumpAllH245(ostream & output,
                 TcpStreamDict & tcpStream,
                 const TcpStreamKey & key,
                 BOOL callerToReceiver)
{
  if (!tcpStream.Contains(key))
    return;

  if (!tcpStream[key].HasTPKT())
    return;

  const char * direction = callerToReceiver ? " (caller -> receiver) "
                                            : " (receiver -> caller) ";

  PPER_Stream per = tcpStream[key].GetTPKT();

  H245_MultimediaSystemControlMessage pdu;
  if (!pdu.Decode(per)) {
    output << "Decode H.245" << direction << " failed.";
    return;
  }

  output << "H.245" << direction << " PDU <" << per.GetSize() << " bytes>:\n  "
         << setprecision(2) << pdu << endl;
}


///////////////////////////////////////////////////////////////////////////////

void DumpAllT120(ostream & output,
                 TcpStreamDict & tcpStream,
                 TcpStreamKey & key,
                 BOOL callerToReceiver)
{
  if (!tcpStream.Contains(key))
    return;

  if (!tcpStream[key].HasTPKT())
    return;

  const char * direction = callerToReceiver ? " (caller -> receiver) "
                                            : " (receiver -> caller) ";

  X224 x224_pdu;
  if (!x224_pdu.Decode(tcpStream[key].GetTPKT())) {
    output << "Read X.224" << direction << "failed.\n";
    return;
  }

  output << "X.224" << direction << " PDU:\n" << setprecision(2) << x224_pdu << endl;

  // An X224 Data PDU...
  if (x224_pdu.GetCode() != X224::DataPDU)
    return;

  key.pdu_count++;

  if (key.pdu_count > 1) {
    PPER_Stream per = x224_pdu.GetData();

    MCS_DomainMCSPDU dom_pdu;
    if (dom_pdu.Decode(per)) {
      output << "T.120" << direction << " Domain PDU:\n  "
             << setprecision(2) << dom_pdu << endl;
      return;
    }
  }
  else {
    PBER_Stream ber = x224_pdu.GetData();

    MCS_ConnectMCSPDU con_pdu;
    if (con_pdu.Decode(ber)) {
      output << "T.120" << direction << " Connect PDU:\n  "
             << setprecision(2) << con_pdu << endl;
      return;
    }
  }

  output << "Decode T.120" << direction << " failed.";
}


///////////////////////////////////////////////////////////////////////////////

void DumpAll(istream & input, ostream & output)
{
  TcpStreamDict tcpStream;

  TcpStreamKey callerSignalChannel;
  TcpStreamKey receiverSignalChannel;
  TcpStreamKey callerControlChannel;
  TcpStreamKey receiverControlChannel;
  TcpStreamList callerConferenceChannels;
  TcpStreamList receiverConferenceChannels;

  while (!input.eof()) {

    TcpStreamKey key;
    PBYTEArray payload;
    BOOL isUDP;
    if (!ReadPacket(input, key, payload, isUDP))
      continue;

    if (isUDP) {
      continue;
    }

    if (!tcpStream.Contains(key))
      tcpStream.SetAt(key, new TcpStreamData);
    tcpStream[key].Concatenate(payload);

    // Look for caller to receiver signal channel
    if (callerSignalChannel.rx_port == 0 && key.rx_port == 1720) {
      callerSignalChannel = key;
      receiverSignalChannel.ReverseCopy(key);
    }

    // Look for caller to receiver control channel
    if (callerControlChannel.tx_port == 0 &&
        key.rx_ip == receiverControlChannel.tx_ip &&
        key.rx_port == receiverControlChannel.tx_port) {
      callerControlChannel = key;
      receiverControlChannel.ReverseCopy(key);
    }

    // Look for caller to receiver conference (T.120) channel
    if (key.rx_port == 1503 &&
                  callerConferenceChannels.GetValuesIndex(key) == P_MAX_INDEX) {
      callerConferenceChannels.Append(new TcpStreamKey(key));
      TcpStreamKey * tmp = new TcpStreamKey;
      tmp->ReverseCopy(key);
      receiverConferenceChannels.Append(tmp);
    }


    // Dump out the Q.931 PDU's in caller signalling channel
    DumpAllQ931(output, tcpStream, callerSignalChannel, TRUE, receiverControlChannel);
    DumpAllQ931(output, tcpStream, receiverSignalChannel, FALSE, receiverControlChannel);

    // Dump out the H.245 PDU's in control channel
    DumpAllH245(output, tcpStream, callerControlChannel, TRUE);
    DumpAllH245(output, tcpStream, receiverControlChannel, FALSE);

    // Dump out the T.120 PDU's in conference channel
    PINDEX i;
    for (i = 0; i < callerConferenceChannels.GetSize(); i++)
      DumpAllT120(output, tcpStream, callerConferenceChannels[i], TRUE);
    for (i = 0; i < receiverConferenceChannels.GetSize(); i++)
      DumpAllT120(output, tcpStream, receiverConferenceChannels[i], FALSE);
  }
}


///////////////////////////////////////////////////////////////////////////////

void DumpOneH245(istream & input, ostream & output)
{
  TcpStreamData data;
  input >> data;

  PPER_Stream per = data;

  H245_MultimediaSystemControlMessage pdu;
  if (!pdu.Decode(per))
    output << "Decode failed!" << endl;

  output << "H.245 PDU:\n  " << setprecision(2) << pdu << endl;

  // The following can be used for debugging purposes, to check out asner.cxx
  per.BeginEncoding();
  pdu.Encode(per);
  per.CompleteEncoding();
}


///////////////////////////////////////////////////////////////////////////////

void DumpOneH225(istream & input, ostream & output)
{
  TcpStreamData data;
  input >> data;

  PPER_Stream per = data;

  H225_H323_UserInformation pdu;
  if (!pdu.Decode(per))
    output << "Decode failed!" << endl;

  output << "H.225 PDU:\n  " << setprecision(2) << pdu << endl;
}


///////////////////////////////////////////////////////////////////////////////

void DumpOneQ931(istream & input, ostream & output)
{
  TcpStreamData data;
  input >> data;

  Q931withH225 q931;
  output << "Q.931:\n  " << setprecision(2) << q931 << endl;
}


///////////////////////////////////////////////////////////////////////////////

PCREATE_PROCESS(DumpPDU);


DumpPDU::DumpPDU()
  : PProcess("Equivalence", "Dump H.323", 1, 0, AlphaCode, 1)
{
}


void DumpPDU::Main()
{
  cout << GetName()
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ')' << endl;

  PArgList & args = GetArguments();
  args.Parse("-q931.-h225.-h245.o:");

  if (args.GetCount() == 0) {
    cerr << "Usage : " << GetFile().GetFileName() << " [options] file [ file ... ]\n"
            "Options:\n"
            "  --q931    : Assume hex is Q.931 PDU(s)\n"
            "  --h225    : Assume hex is H.225 PDU(s)\n"
            "  --h245    : Assume hex is H.245 PDU(s)\n"
            "  -o file   : Output dump to file (default stdout)\n"
            "  file      : Hex file (use '-' for stdin)\n"
            "\n"
            "If none of the protocol options above is selected then the input is assumed to\n"
            "be the output from the command `tcpdump -nx -s 1500 ip host <host1> and <host2>'\n"
            "which captures all IP data between the two hosts.\n";
    return;
  }

  PTextFile outfile;
  ostream * output = &cout;
  if (args.HasOption('o')) {
    output = &outfile;
    if (!outfile.Open(args.GetOptionString('o'), PFile::WriteOnly)) {
      cerr << GetName() << ": Could not open '"
           << outfile.GetFilePath() << "' - " << outfile.GetErrorText() << endl;
      return;
    }
  }

  for (PINDEX arg = 0; arg < args.GetCount(); arg++) {
    PTextFile infile;
    istream * input = &cin;
    if (args[arg] != "-") {
      input = &infile;
      if (!infile.Open(args[arg], PFile::ReadOnly)) {
        cerr << GetName() << ": Could not open '"
             << infile.GetFilePath() << "' - " << infile.GetErrorText() << endl;
        continue;
      }
    }

    if (args.HasOption("q931"))
      DumpOneQ931(*input, *output);
    else if (args.HasOption("h225"))
      DumpOneH225(*input, *output);
    else if (args.HasOption("h245"))
      DumpOneH245(*input, *output);
    else
      DumpAll(*input, *output);
  }
}


// End of File ///////////////////////////////////////////////////////////////
