/*
 * main.cxx
 *
 * PWLib application source file for OpenPhone
 *
 * Main program entry point.
 *
 * Copyright 1999 Equivalence
 *
 * $Log: main.cxx,v $
 * Revision 1.2  1999/09/01 06:12:21  robertj
 * Added gatekeeper support.
 *
 * Revision 1.1  1999/08/25 05:16:11  robertj
 * GUI test application.
 *
 * Revision 1.2  1999/01/26 06:35:07  robertj
 * Fixed $LOG$ variable in template files
 *
 */

#include <pwlib.h>
#include "main.h"
#include "resources.h"


PCREATE_PROCESS(OpenPhone);

#define new PNEW


///////////////////////////////////////////////////////////////////////////////

OpenPhone::OpenPhone()
  : PApplication("Equivalence", "OpenPhone", 1, 0, AlphaCode, 1)
{
}


void OpenPhone::Main()
{
  // initialise the About dialog
  SetAboutDialogID(IDD_ABOUT);

  // create the main window and endpoint
  new MainWindow(GetArguments());

  // enter the main loop
  PApplication::Main();
}


///////////////////////////////////////////////////////////////////////////////

MainWindow::MainWindow(PArgList & /*args*/)
  : PTopLevelWindow(CanClose|CanIconify|CanZoom)
{
  SetTitle(PResourceString(IDS_TITLE));
  SetIcon(PIcon(IDI_MAIN_WINDOW));
  SetMenu(new MainMenu());

  // set the background colour and load the image
  SetBackgroundColour(PColour::Grey);

  // set the window size
  SetDimensions(200, 100, LocalCoords);

  // create status window
  statusWindow = new PStaticText(this);
  statusWindow->SetDimensions(190, 20, LocalCoords);
  statusWindow->SetPosition  (5, 5);
  statusWindow->SetBackgroundColour(PColour::Black);
  statusWindow->SetForegroundColour(PColour::Green);

  // create Call/Answer button
  callButton = new PTextButton(this, PResourceString(IDS_CALL), PCREATE_NOTIFIER(OnCallButton));
  callButton->SetDimensions(100, 12, LocalCoords);
  callButton->SetPosition  (5, 27);

  //PEditBox * editBox = new PEditBox(this);
  //editBox->SetDimensions(120, 10, LocalCoords);
  //editBox->SetPosition  ( 50, 10);

  //display the window
  UpdateCommandSources();
  ShowAll();

  // create H323 endpoint
  endPoint = new OpenPhoneEndPoint(this);

  // initialise the endpoint
  PConfig config(PConfig::Application);
  endPoint->Initialise(config);

  // start the H323 stack listening for incoming calls
  H323ListenerTCP * listener = new H323ListenerTCP(*endPoint);
  if (!endPoint->StartListener(listener)) {
    PSimpleDialog::Error(this, IDS_NOPORT, listener->GetListenerPort());
    callState = Bogus;
  }

  // set the internal state
  callState = Listening;
  SetStatus(PResourceString(IDS_LISTENING));
}


MainWindow::~MainWindow()
{
  delete endPoint;
}


void MainWindow::OnMenuStartSelect()
{
  PSubMenu & rootMenu = *GetMenu();

  // get handles to menus
  PSubMenu & callMenu = (PSubMenu &)rootMenu[0];

  // enable/disable options
  PMenuItem & acceptCallItem = (PMenuItem &)callMenu[2];
  acceptCallItem.Disable();

  PMenuItem & refuseCallItem = (PMenuItem &)callMenu[3];
  refuseCallItem.Disable();
}


void MainWindow::OnCallButton(PTextButton &, INT)
{
  switch (callState) {
    case Listening:
      CallCmd();
      break;

    case Calling:
    case InCall:
      HangupCmd();
      break;

    default:
      break;
  }
}


void MainWindow::OnConnectionEstablished(const PString & token)
{
  currentCallToken = token;
  switch (callState) {
    case Calling:
    case Answering:
      SetStatus(psprintf(PResourceString(IDS_TALKING), (const char *)remoteParty));
      callState = InCall;
      callButton->SetName(PResourceString(IDS_HANGUP));
      break;

    default:
      PSimpleDialog::Info(this, "Connection established - %i", callState);
      break;
  }
}


void MainWindow::OnConnectionCleared()
{
  switch (callState) {
    case Calling:
      SetStatus(psprintf(PResourceString(IDS_REFUSED), (const char *)remoteParty)  + "\n" + PResourceString(IDS_LISTENING));
      break;

    case InCall:
      SetStatus(psprintf(PResourceString(IDS_REMOTEHUNGUP), (const char *)remoteParty)  + "\n" + PResourceString(IDS_LISTENING));
      break;

    case HangingUp:
      SetStatus(psprintf(PResourceString(IDS_HUNGUP), (const char *)remoteParty)  + "\n" + PResourceString(IDS_LISTENING));
      break;

    default:
      PSimpleDialog::Info(this, "Connection cleared - %i", callState);
      break;
  }

  callState = Listening;
  callButton->SetName(PResourceString(IDS_CALL));
  callButton->Enable();
}


BOOL MainWindow::OnAnswerCall(const PString & caller)
{
  if (callState != Listening)
    return FALSE;

  BOOL answer = PSimpleDialog::YesNo(this, IDS_ASKACCEPT, (const char *)caller);

  if (answer) {
    callState = Answering;
    remoteParty = caller;
  }

  return answer;
}


BOOL MainWindow::OnAlerting(const PString & /*caller*/)
{
  return TRUE;
}


void MainWindow::HangupCmd()
{
  endPoint->ClearCall(currentCallToken);
  callState = HangingUp;
  SetStatus(PResourceString(IDS_HANGINGUP));
  callButton->Disable();
}


void MainWindow::CallCmd()
{
  MakeCallDialog dlg(this);

  // ask the user what address to call
  if (dlg.RunModal()) {
    SetStatus(psprintf(PResourceString(IDS_CALLING), (const char *)dlg.address));
    callButton->SetName(PResourceString(IDS_HANGUP));
    callState = Calling;

    H323TransportTCP * transport = new H323TransportTCP(*endPoint);
    if (!dlg.gateway) {
      if (!transport->SetRemoteAddress("ip:"+dlg.gateway)) {
        //cout << "Invalid gateway address: \"" << dlg.gateway << '"' << endl;
        return;
      }
    }
    else {
      if (!transport->SetRemoteAddress("ip:"+dlg.address)) {
        //cout << "Invalid destination address: \"" << dlg.address << '"' << endl;
        return;
      }
    }

    endPoint->MakeCall(dlg.address, transport, currentCallToken);
  }
}


void MainWindow::AcceptCallCmd()
{
}


void MainWindow::RefuseCallCmd()
{
}


void MainWindow::ExitCmd()
{
  owner->Terminate();
}


//void MainWindow::AutoAnswerCmd()
//{
//}


//void MainWindow::UseGatewayCmd()
//{
//}


//void MainWindow::ConfigCodecsCmd()
//{
//}


///////////////////////////////////////////////////////////////////////////////

void MakeCallDialog::AddressChanged(PEditBox & box, INT)
{
  ok->Enable(box.GetLength() > 0);
}


///////////////////////////////////////////////////////////////////////////////

OpenPhoneConnection::OpenPhoneConnection(OpenPhoneEndPoint & endpoint,
                                         unsigned callReference,
                                         MainWindow * _mainWindow)
  : H323Connection(endpoint, callReference),
    mainWindow(_mainWindow)
{
}

BOOL OpenPhoneConnection::OnAnswerCall(const PString & caller,
                                       const H323SignalPDU & /*setupPDU*/,
                                             H323SignalPDU & /*connectPDU*/)
{
  return mainWindow->OnAnswerCall(caller);
}


BOOL OpenPhoneConnection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                     const PString & username)
{
  return mainWindow->OnAlerting(username);
}


///////////////////////////////////////////////////////////////////////////////

OpenPhoneEndPoint::OpenPhoneEndPoint(MainWindow * _mainWindow)
  : mainWindow(_mainWindow)
{
}


void OpenPhoneEndPoint::OnConnectionEstablished(H323Connection &, const PString & token)
{
  mainWindow->OnConnectionEstablished(token);
}


void OpenPhoneEndPoint::OnConnectionCleared(H323Connection &, const PString &)
{
  mainWindow->OnConnectionCleared();
}


H323Connection * OpenPhoneEndPoint::CreateConnection(unsigned callReference)
{
  return new OpenPhoneConnection(*this, callReference, mainWindow);
}


void OpenPhoneEndPoint::Initialise(PConfig & config)
{
  // set record and play devices
  config.SetDefaultSection("Devices");
  SetSoundChannelPlayDevice  (config.GetString("soundplay",   GetSoundChannelPlayDevice()));
  SetSoundChannelRecordDevice(config.GetString("soundrecord", GetSoundChannelRecordDevice()));

  // create the capabilities
  H323Capability * gsm    = new H323_GSM0610Capability;
  H323Capability * ulaw64 = new H323_G711Capability(H323_G711Capability::muLaw, H323_G711Capability::At64k);
  H323Capability * alaw64 = new H323_G711Capability(H323_G711Capability::ALaw, H323_G711Capability::At64k);
  //H323Capability * t120   = new H323_T120Capability;
  H323Capability * uic    = new H323_UserInputCapability;

  // add the capabilities in the desired order
  AddCapability(gsm);
  AddCapability(ulaw64);
  AddCapability(alaw64);
  //AddCodec(t120);
  AddCapability(uic);

  // now set them
  SetCapability(0, 0, gsm);
  SetCapability(0, 0, ulaw64);
  SetCapability(0, 0, alaw64);
  //SetCapability(0, 1, t120);
  SetCapability(0, 0, uic);

  config.SetDefaultSection("Options");
  SetLocalUserName(config.GetString("username", GetLocalUserName()));
  SetMaxAudioDelayJitter(config.GetInteger("jitter",  100));
}


// End of File ///////////////////////////////////////////////////////////////
