/*
 * h323.cxx
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: h323.cxx,v $
 * Revision 1.43  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.42  1999/08/27 15:42:44  craigs
 * Fixed problem with local call tokens using ambiguous interface names, and connect timeouts not changing connection state
 *
 * Revision 1.41  1999/08/25 05:11:22  robertj
 * File fission (critical mass reached).
 * Improved way in which remote capabilities are created, removed case statement!
 * Changed MakeCall, so immediately spawns thread, no black on TCP connect.
 *
 * Revision 1.40  1999/08/14 03:26:45  robertj
 * Compiler compatibility problems.
 *
 * Revision 1.39  1999/08/10 04:20:26  robertj
 * Fixed constness problems in some PASN_Choice casts.
 *
 * Revision 1.38  1999/08/08 10:03:33  robertj
 * Fixed capability selection to honor local table priority order.
 *
 * Revision 1.37  1999/07/23 02:37:53  robertj
 * Fixed problems with hang ups and crash closes of connections.
 *
 * Revision 1.36  1999/07/19 02:01:02  robertj
 * Fixeed memory leask on connection termination.
 * Fixed connection "orderly close" on endpoint termination.
 *
 * Revision 1.35  1999/07/18 14:57:29  robertj
 * Fixed bugs in slow start with H245 tunnelling, part 4.
 *
 * Revision 1.34  1999/07/18 14:29:31  robertj
 * Fixed bugs in slow start with H245 tunnelling, part 3.
 *
 * Revision 1.33  1999/07/18 13:59:12  robertj
 * Fixed bugs in slow start with H245 tunnelling, part 2.
 *
 * Revision 1.32  1999/07/18 13:25:40  robertj
 * Fixed bugs in slow start with H245 tunnelling.
 *
 * Revision 1.31  1999/07/18 10:19:39  robertj
 * Fixed CreateCapability function: missing break's in case!
 *
 * Revision 1.30  1999/07/17 06:22:32  robertj
 * Fixed bug in setting up control channel when initiating call (recently introduced)
 *
 * Revision 1.29  1999/07/16 16:04:41  robertj
 * Bullet proofed local capability table data entry.
 *
 * Revision 1.28  1999/07/16 14:04:47  robertj
 * Fixed bug that caused signal channel to be closed, forgot to disable connect time.
 *
 * Revision 1.27  1999/07/16 06:15:59  robertj
 * Corrected semantics for tunnelled master/slave determination in fast start.
 *
 * Revision 1.26  1999/07/16 02:15:20  robertj
 * Fixed more tunneling problems.
 * Fixed fastStart initiator matching response to correct channels.
 *
 * Revision 1.25  1999/07/16 00:51:03  robertj
 * Some more debugging of fast start.
 *
 * Revision 1.24  1999/07/15 14:45:36  robertj
 * Added propagation of codec open error to shut down logical channel.
 * Fixed control channel start up bug introduced with tunnelling.
 *
 * Revision 1.23  1999/07/15 09:04:31  robertj
 * Fixed some fast start bugs
 *
 * Revision 1.22  1999/07/14 06:06:14  robertj
 * Fixed termination problems (race conditions) with deleting connection object.
 *
 * Revision 1.21  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.20  1999/07/13 02:50:58  craigs
 * Changed semantics of SetPlayDevice/SetRecordDevice, only descendent
 *    endpoint assumes PSoundChannel devices for audio codec.
 *
 * Revision 1.19  1999/07/10 02:59:26  robertj
 * Fixed ability to hang up incoming connection.
 *
 * Revision 1.18  1999/07/10 02:51:36  robertj
 * Added mutexing in H245 procedures. Also fixed MSD state bug.
 *
 * Revision 1.17  1999/07/09 14:59:59  robertj
 * Fixed GNU C++ compatibility.
 *
 * Revision 1.16  1999/07/09 06:09:49  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.15  1999/06/25 10:25:35  robertj
 * Added maintentance of callIdentifier variable in H.225 channel.
 *
 * Revision 1.14  1999/06/22 13:45:05  robertj
 * Added user question on listener version to accept incoming calls.
 *
 * Revision 1.13  1999/06/14 06:39:08  robertj
 * Fixed problem with getting transmit flag to channel from PDU negotiator
 *
 * Revision 1.12  1999/06/14 05:15:55  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 * Revision 1.11  1999/06/13 12:41:14  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.10  1999/06/09 06:18:00  robertj
 * GCC compatibiltiy.
 *
 * Revision 1.9  1999/06/09 05:26:19  robertj
 * Major restructuring of classes.
 *
 * Revision 1.8  1999/06/07 00:37:05  robertj
 * Allowed for reuseaddr on listen when in debug version
 *
 * Revision 1.7  1999/06/06 06:06:36  robertj
 * Changes for new ASN compiler and v2 protocol ASN files.
 *
 * Revision 1.6  1999/04/26 06:20:22  robertj
 * Fixed bugs in protocol
 *
 * Revision 1.5  1999/04/26 06:14:46  craigs
 * Initial implementation for RTP decoding and lots of stuff
 * As a whole, these changes are called "First Noise"
 *
 * Revision 1.4  1999/02/23 11:04:28  robertj
 * Added capability to make outgoing call.
 *
 * Revision 1.3  1999/01/16 01:31:33  robertj
 * Major implementation.
 *
 * Revision 1.2  1999/01/02 04:00:56  robertj
 * Added higher level protocol negotiations.
 *
 * Revision 1.1  1998/12/14 09:13:09  robertj
 * Initial revision
 *
 */

#include <ptlib.h>
#include "h323.h"
#include "h323neg.h"
#include "channels.h"


/////////////////////////////////////////////////////////////////////////////

#ifdef PTRACING
const char * const H323Connection::FastStartStateNames[NumFastStartStates] = {
  "FastStartDisabled", "FastStartInitiate", "FastStartResponse"
};
#endif


H323Connection::H323Connection(H323EndPoint & end, unsigned ref)
  : endpoint(end)
{
  callAnswered = FALSE;
  callReference = ref;

  signallingChannel = NULL;
  controlChannel = NULL;
  h245Tunneling = TRUE;
  h245TunnelPDU = NULL;

  connectionState = NoConnectionActive;

  bandwidthAvailable = endpoint.GetInitialBandwidth();

  fastStartState = FastStartDisabled;

  remoteMaxAudioDelayJitter = 0;

  masterSlaveDeterminationProcedure = new H245NegMasterSlaveDetermination(endpoint, *this);
  capabilityExchangeProcedure = new H245NegTerminalCapabilitySet(endpoint, *this);
  logicalChannels = new H245NegLogicalChannels(endpoint, *this);
  requestModeProcedure = new H245NegRequestMode(endpoint, *this);
  roundTripDelayProcedure = new H245NegRoundTripDelay(endpoint, *this);
}


H323Connection::~H323Connection()
{
  delete signallingChannel;
  delete controlChannel;

  delete masterSlaveDeterminationProcedure;
  delete capabilityExchangeProcedure;
  delete logicalChannels;
  delete requestModeProcedure;
  delete roundTripDelayProcedure;

  PTRACE(2, "H323\tConnection finished.");
}


void H323Connection::Close(BOOL gkForced)
{
  static PMutex closeMutex;

  closeMutex.Wait();

  switch (connectionState) {
    case NoConnectionActive :
    case ShuttingDownConnection :
      closeMutex.Signal();
      return;

    case AwaitingTransportConnect :
      signallingChannel->Close();
      connectionState = NoConnectionActive;
      closeMutex.Signal();
      endpoint.SignalCleanUpConnections();
      return;

    default :
      connectionState = ShuttingDownConnection;
      closeMutex.Signal();
      break;
  }

  if (controlChannel != NULL || (h245Tunneling && signallingChannel != NULL)) {
    H323ControlPDU pdu;
    pdu.BuildEndSessionCommand(H245_EndSessionCommand::e_disconnect);
    WriteControlPDU(pdu);

    // This will close the control channel and wait for its thread to terminate
    if (controlChannel != NULL)
      controlChannel->CloseControlChannel();
  }

  // This will close the signalling channel (sending a ReleaseComplete) and
  // wait for its thread to terminate
  if (signallingChannel != NULL)
    signallingChannel->CloseSignallingChannel(callReference, callIdentifier);

  // Check for gatekeeper and do admission check if have one
  if (!gkForced) {
    H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
    if (gatekeeper != NULL)
      gatekeeper->DisengageRequest(*this, H225_DisengageReason::e_normalDrop);
  }

  // Clear flag so clean up thread disposes of the connection
  connectionState = NoConnectionActive;

  endpoint.SignalCleanUpConnections();
}


BOOL H323Connection::IsOpen() const
{
  return connectionState == ShuttingDownConnection ||
         connectionState == AwaitingTransportConnect ||
         (controlChannel != NULL && controlChannel->IsOpen()) ||
         (signallingChannel != NULL && signallingChannel->IsOpen());
}


void H323Connection::AttachSignalChannel(H323Transport * channel)
{
  if (signallingChannel != NULL && signallingChannel->IsOpen()) {
    PAssertAlways(PLogicError);
    return;
  }

  delete signallingChannel;
  signallingChannel = channel;
}


BOOL H323Connection::WriteSignalPDU(H323SignalPDU & pdu)
{
  PAssert(signallingChannel != NULL, PLogicError);
  if (signallingChannel == NULL)
    return FALSE;

  pdu.m_h323_uu_pdu.m_h245Tunneling = h245Tunneling;

  return pdu.Write(*signallingChannel);
}


void H323Connection::HandleSignallingChannel()
{
  PAssert(signallingChannel != NULL, PLogicError);

  PTRACE(2, "H225\tReading PDUs: callRef=" << callReference);

  H323SignalPDU pdu;
  while (signallingChannel->IsOpen()) {
    if (pdu.Read(*signallingChannel) && !HandleSignalPDU(pdu)) {
      signallingChannel->CloseSignallingChannel(callReference, callIdentifier);
      break;
    }
  }

  PTRACE(2, "H225\tSignal channel closed.");

  Close();
}


BOOL H323Connection::HandleSignalPDU(H323SignalPDU & pdu)
{
  const Q931 & q931 = pdu.GetQ931();

  PTRACE(3, "H225\tHandling PDU: " << q931.GetMessageTypeName()
                    << " callRef=" << q931.GetCallReference());

  // If remote does not do tunneling, so we don't either. Note that if it
  // gets turned off once, it stays off for good.
  h245Tunneling = h245Tunneling && pdu.m_h323_uu_pdu.m_h245Tunneling;

  switch (q931.GetMessageType()) {
    case Q931::SetupMsg :
      if (!OnReceivedSignalSetup(pdu))
        return FALSE;
      break;

    case Q931::CallProceedingMsg :
      if (!OnReceivedCallProceeding(pdu))
        return FALSE;
      break;

    case Q931::AlertingMsg :
      if (!OnReceivedAlerting(pdu))
        return FALSE;
      break;

    case Q931::ConnectMsg :
      if (!OnReceivedSignalConnect(pdu))
        return FALSE;
      break;

    case Q931::FacilityMsg :
      if (!OnReceivedFacility(pdu))
        return FALSE;
      break;

    case Q931::ReleaseCompleteMsg :
      return FALSE;

    default :
      if (!OnUnknownSignalPDU(pdu))
        return FALSE;
  }

  // Process tunnelled H245 PDU, if present.
  HandleTunnelPDU(pdu, NULL);
  return TRUE;
}


void H323Connection::HandleTunnelPDU(const H323SignalPDU & pdu, H323SignalPDU * reply)
{
  if (!h245Tunneling)
    return;

  h245TunnelPDU = reply;

  for (PINDEX i = 0; i < pdu.m_h323_uu_pdu.m_h245Control.GetSize(); i++) {
    PPER_Stream strm = pdu.m_h323_uu_pdu.m_h245Control[i].GetValue();
    HandleControlData(strm);
  }

  h245TunnelPDU = NULL;

  // Make sure does not get repeated, break const and clear tunnelled H.245 PDU's
  ((H323SignalPDU &)pdu).m_h323_uu_pdu.m_h245Control.SetSize(0);
}


static BOOL BuildFastStartList(const H323Channel & channel,
                               H225_ArrayOf_PASN_OctetString & array,
                               H323Channel::Directions reverseDirection)
{
  H245_OpenLogicalChannel open;
  const H323Capability & capability = channel.GetCapability();

  if (channel.GetDirection() != reverseDirection) {
    if (!capability.OnSendingPDU(open.m_forwardLogicalChannelParameters.m_dataType))
      return FALSE;
  }
  else {
    if (!capability.OnSendingPDU(open.m_reverseLogicalChannelParameters.m_dataType))
      return FALSE;

    open.m_forwardLogicalChannelParameters.m_multiplexParameters.SetTag(
                H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters::e_none);
    open.m_forwardLogicalChannelParameters.m_dataType.SetTag(H245_DataType::e_nullData);
    open.IncludeOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters);
  }

  if (!channel.OnSendingPDU(open))
    return FALSE;

  PTRACE(4, "H225\tBuild fastStart:\n  " << setprecision(2) << open);
  PPER_Stream strm;
  open.Encode(strm);
  strm.CompleteEncoding();

  PINDEX last = array.GetSize();
  array.SetSize(last+1);
  array[last] = strm;
  return TRUE;
}


BOOL H323Connection::OnReceivedSignalSetup(const H323SignalPDU & setupPDU)
{
  PINDEX i;

  const H225_Setup_UUIE & setup = setupPDU.m_h323_uu_pdu.m_h323_message_body;

  // Indicate the direction of call.
  callAnswered = TRUE;

  // Set our call token for identification in endpoint dictionary
  callToken = H323EndPoint::BuildConnectionToken(*signallingChannel, callReference, TRUE);

  // Save the identifiers sent by caller
  if (setup.HasOptionalField(H225_Setup_UUIE::e_callIdentifier))
    callIdentifier = setup.m_callIdentifier.m_guid;
  conferenceIdentifier = setup.m_conferenceID;

  if (setup.HasOptionalField(H225_Setup_UUIE::e_sourceAddress) &&
      setup.m_sourceAddress.GetSize() > 0) {
    remotePartyName = H323GetAliasAddressString(setup.m_sourceAddress[0]);
    if (setup.m_sourceAddress.GetSize() > 1) {
      remotePartyName += " (";
      for (i = 1; i < setup.m_sourceAddress.GetSize(); i++) {
        if (i > 1)
          remotePartyName += ", ";
        remotePartyName += H323GetAliasAddressString(setup.m_sourceAddress[i]);
      }
      remotePartyName += ')';
    }
  }
  else {
    remotePartyName = setupPDU.GetQ931().GetDisplayName();
    if (remotePartyName.IsEmpty())
      remotePartyName = signallingChannel->GetRemoteAddress();
  }

  // See if remote endpoint wants to start fast
  if (setup.HasOptionalField(H225_Setup_UUIE::e_fastStart)) {
    remoteCapabilityTable.RemoveAll();
    PTRACE(3, "H225\tFast start detected");

    // Extract capabilities from the fast start OpenLogicalChannel structures
    for (i = 0; i < setup.m_fastStart.GetSize(); i++) {
      H245_OpenLogicalChannel open;
      PPER_Stream strm = (PBYTEArray)setup.m_fastStart[i];
      if (open.Decode(strm)) {
        PTRACE(4, "H225\tFast start open:\n  " << setprecision(2) << open);
        unsigned error;
        H323Channel * channel = endpoint.CreateLogicalChannel(*this, open, error);
        if (channel != NULL)
          fastStartChannels.Append(channel);
      }
    }

    // If we are incapable of ANY of the fast start channels, don't do fast start
    if (!fastStartChannels.IsEmpty())
      fastStartState = FastStartResponse;
  }

  // send back a H323 Call Proceeding PDU in case OnIncomingCall() takes a while
  PTRACE(3, "H225\tSending call proceeding PDU");
  H323SignalPDU callProceedingPDU;
  callProceedingPDU.BuildCallProceeding(endpoint, callReference, callIdentifier);
  if (h245Tunneling) {
    // Piggy back H245 on this reply
    h245TunnelPDU = &callProceedingPDU;
    if (fastStartState != FastStartResponse)
      capabilityExchangeProcedure->Start();
    HandleTunnelPDU(setupPDU, &callProceedingPDU);
  }
  if (!WriteSignalPDU(callProceedingPDU))
    return FALSE;

  // Check for gatekeeper and do admission check if have one
  H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
  if (gatekeeper != NULL) {
    if (!gatekeeper->AdmissionRequest(*this))
      return FALSE;
  }

  // if the application indicates not to contine, then send a Q931 Release Complete PDU
  H323SignalPDU alertingPDU;
  alertingPDU.BuildAlerting(endpoint, callReference, callIdentifier);
  if (!OnIncomingCall(setupPDU, alertingPDU)) {
    PTRACE(1, "H225\tApplication not accepting calls");
    return FALSE;
  }

  // send Q931 Alerting PDU
  PTRACE(3, "H225\tIncoming call accepted - sending Alerting PDU");
  HandleTunnelPDU(setupPDU, &alertingPDU);
  if (!WriteSignalPDU(alertingPDU))
    return FALSE;

  // Build the reply with the channels we are actually using
  H323SignalPDU connectPDU;
  H225_Connect_UUIE & connect = connectPDU.BuildConnect(endpoint,
                                                        callReference,
                                                        callIdentifier,
                                                        conferenceIdentifier);

  // call the application callback to determine if to answer the call or not
  if (!OnAnswerCall(remotePartyName, setupPDU, connectPDU)) {
    PTRACE(1, "H225\tApplication has declined to answer incoming call");
    return FALSE;
  }

  // Now ask the application to select which channels to start
  if (fastStartState == FastStartResponse)
    OnSelectLogicalChannels();

  // Remove any channels that were not started by OnSelectLogicalChannels(),
  // those that were started are put into the logical channel dictionary
  for (i = 0; i < fastStartChannels.GetSize(); i++) {
    if (fastStartChannels[i].IsRunning())
      logicalChannels->Add(fastStartChannels[i]);
    else
      fastStartChannels.RemoveAt(i--);
  }

  // The channels we just transferred to the logical channels dictionary
  // should not be deleted via this structure now.
  fastStartChannels.DisallowDeleteObjects();

  // Set flag that we are up to CONNECT stage
  if (connectionState == ShuttingDownConnection)
    return FALSE;
  connectionState = HasExecutedSignalConnect;

  // No channels selected (or never provided) so do traditional H245 start
  if (fastStartChannels.IsEmpty()) {
    if (h245Tunneling) {
      HandleTunnelPDU(setupPDU, &connectPDU);
      return WriteSignalPDU(connectPDU); // Send H323 Connect PDU
    }
    controlChannel = signallingChannel->CreateControlChannel(*this);
    return controlChannel != NULL;
  }

  PTRACE(3, "H225\tAccepting fastStart for " << fastStartChannels.GetSize() << " channels");

  connect.IncludeOptionalField(H225_Connect_UUIE::e_fastStart);

  for (i = 0; i < fastStartChannels.GetSize(); i++)
    BuildFastStartList(fastStartChannels[i], connect.m_fastStart, H323Channel::IsTransmitter);

  // If have some acceptable channels, do the fast start
  if (connect.m_fastStart.GetSize() > 0) {
    endpoint.OnConnectionEstablished(*this, callToken);
    return WriteSignalPDU(connectPDU); // Send H323 Connect PDU
  }

  // Otherwise something really odd just happened, abort call
  PTRACE(1, "H225\tFast start error!");
  return FALSE;
}


BOOL H323Connection::OnReceivedCallProceeding(const H323SignalPDU & pdu)
{
  const H225_CallProceeding_UUIE & call = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (call.HasOptionalField(H225_CallProceeding_UUIE::e_fastStart))
    return HandleFastStartAcknowledge(call.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (call.HasOptionalField(H225_CallProceeding_UUIE::e_h245Address))
    return StartControlChannel(call.m_h245Address);

  return TRUE;
}


BOOL H323Connection::OnReceivedAlerting(const H323SignalPDU & pdu)
{
  const H225_Alerting_UUIE & alert = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (alert.HasOptionalField(H225_Alerting_UUIE::e_fastStart))
    return HandleFastStartAcknowledge(alert.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (alert.HasOptionalField(H225_Alerting_UUIE::e_h245Address))
    return StartControlChannel(alert.m_h245Address);

  remotePartyName = pdu.GetQ931().GetDisplayName();
  if (remotePartyName.IsEmpty())
    remotePartyName = signallingChannel->GetRemoteAddress();

  return OnAlerting(pdu, remotePartyName);
}


BOOL H323Connection::OnReceivedSignalConnect(const H323SignalPDU & pdu)
{
  if (connectionState == ShuttingDownConnection)
    return FALSE;
  connectionState = HasExecutedSignalConnect;

  if (!OnOutgoingCall(pdu))
    return FALSE;

  const H225_Connect_UUIE & connect = pdu.m_h323_uu_pdu.m_h323_message_body;

  // have answer, so set timeout to not happen
  signallingChannel->SetReadTimeout(PMaxTimeInterval);

  // Check for fastStart data and start fast
  if (connect.HasOptionalField(H225_Connect_UUIE::e_fastStart))
    return HandleFastStartAcknowledge(connect.m_fastStart);

  fastStartState = FastStartDisabled;
  fastStartChannels.RemoveAll();

  // Check that it has the H.245 channel connection info
  if (connect.HasOptionalField(H225_Connect_UUIE::e_h245Address))
    return StartControlChannel(connect.m_h245Address);

  if (!h245Tunneling) {
    PTRACE(2, "H225\tNo H245 address in remote PDU");
    return FALSE;
  }

  if (masterSlaveDeterminationProcedure->IsDetermined() &&
      capabilityExchangeProcedure->HasSentCapabilities() &&
      capabilityExchangeProcedure->HasReceivedCapabilities()) {
    if (!OnControlChannelOpen())
      return FALSE;
    connectionState = EstablishedConnection;
  }

  return TRUE;
}


BOOL H323Connection::OnReceivedFacility(const H323SignalPDU & pdu)
{
  if (pdu.m_h323_uu_pdu.m_h323_message_body.GetTag() ==
                            H225_H323_UU_PDU_h323_message_body::e_empty)
    return TRUE;

  const H225_Facility_UUIE & fac = pdu.m_h323_uu_pdu.m_h323_message_body;

  // Check for fastStart data and start fast
  if (fac.HasOptionalField(H225_Facility_UUIE::e_fastStart))
    return HandleFastStartAcknowledge(fac.m_fastStart);

  // Check that it has the H.245 channel connection info
  if (fac.HasOptionalField(H225_CallProceeding_UUIE::e_h245Address))
    return StartControlChannel(fac.m_h245Address);

  return TRUE;
}


BOOL H323Connection::OnIncomingCall(const H323SignalPDU &, H323SignalPDU &)
{
  return TRUE;
}


BOOL H323Connection::OnAnswerCall(const PString & caller,
                                  const H323SignalPDU &,
                                  H323SignalPDU &)
{
  PTRACE(2, "H225\tOnAnswerCall from \"" << caller << '"');
  return TRUE;
}


BOOL H323Connection::SendSignalSetup(const PString & alias)
{
  connectionState = AwaitingGatekeeperAdmission;

  // Indicate the direction of call.
  callAnswered = FALSE;
  remotePartyName = alias;

  // Start building the setup PDU to get various ID's
  H323SignalPDU setupPDU;
  H225_Setup_UUIE & setup = setupPDU.BuildSetup(endpoint,
                                                endpoint.GetLocalUserName(),
                                                remotePartyName);

  // Save the identifiers generated by BuildSetup
  callToken = H323EndPoint::BuildConnectionToken(*signallingChannel, callReference, FALSE);
  callIdentifier = setup.m_callIdentifier.m_guid;
  conferenceIdentifier = setup.m_conferenceID;

  // Check for gatekeeper and do admission check if have one
  H323Gatekeeper * gatekeeper = endpoint.GetGatekeeper();
  if (gatekeeper != NULL) {
    H323TransportAddress gatekeeperRoute;
    if (!gatekeeper->AdmissionRequest(*this, gatekeeperRoute)) {
      PTRACE(1, "H225\tGatekeeper refused admission.");
      return FALSE;
    }
    if (!gatekeeperRoute && !signallingChannel->SetRemoteAddress(gatekeeperRoute)) {
      PTRACE(1, "H225\tInvalid gatekeeper route address: \"" << gatekeeperRoute << '"');
      connectionState = AwaitingTransportConnect;
      return FALSE;
    }
  }

  // Do the transport connect
  connectionState = AwaitingTransportConnect;
  if (!signallingChannel->Connect()) {
    connectionState = NoConnectionActive;
    return FALSE;
  }

  PTRACE(3, "H225\tSending Setup PDU");
  connectionState = AwaitingSignalConnect;

  // Put in all the signalling addresses for link
  setup.IncludeOptionalField(H225_Setup_UUIE::e_sourceCallSignalAddress);
  setup.m_sourceCallSignalAddress.SetTag(H225_TransportAddress::e_ipAddress);
  setup.IncludeOptionalField(H225_Setup_UUIE::e_sourceCallSignalAddress);
  signallingChannel->SetUpTransportPDU(setup.m_sourceCallSignalAddress, TRUE);
  setup.IncludeOptionalField(H225_Setup_UUIE::e_destCallSignalAddress);
  signallingChannel->SetUpTransportPDU(setup.m_destCallSignalAddress, FALSE);

  // Ask the application what channels to open
  PTRACE(3, "H225\tFast start begun by local endpoint");
  fastStartState = FastStartInitiate;
  fastStartChannels.RemoveAll();
  OnSelectLogicalChannels();

  // If application called OpenLogicalChannel, put in the fastStart field
  if (!fastStartChannels.IsEmpty()) {
    for (PINDEX i = 0; i < fastStartChannels.GetSize(); i++)
      BuildFastStartList(fastStartChannels[i], setup.m_fastStart, H323Channel::IsReceiver);
    if (setup.m_fastStart.GetSize() > 0)
      setup.IncludeOptionalField(H225_Setup_UUIE::e_fastStart);
  }

  if (!OnSendSignalSetup(setupPDU))
    return FALSE;

  fastStartState = FastStartDisabled;

  if (h245Tunneling) {
    h245TunnelPDU = &setupPDU;

    // Try and start the master/slave and capability exchange through the tunnel
    masterSlaveDeterminationProcedure->Start();

    // Cannot do capability exchange if if fast starting though.
    if (fastStartChannels.IsEmpty())
      capabilityExchangeProcedure->Start();

    h245TunnelPDU = NULL;
  }

  // Send the initial PDU
  if (!WriteSignalPDU(setupPDU))
    return FALSE;

  // Set timeout for remote party to answer the call
  signallingChannel->SetReadTimeout(endpoint.GetSignallingChannelCallTimeout());

  return TRUE;
}


BOOL H323Connection::OnSendSignalSetup(H323SignalPDU & /*setupPDU*/)
{
  return TRUE;
}


BOOL H323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                const PString & /*username*/)
{
  PTRACE(1, "H225\tReceived alerting PDU.");
  return TRUE;
}


BOOL H323Connection::OnOutgoingCall(const H323SignalPDU & /*connectPDU*/)
{
  PTRACE(1, "H225\tReceived connect PDU.");
  return TRUE;
}


BOOL H323Connection::HandleFastStartAcknowledge(const H225_ArrayOf_PASN_OctetString & array)
{
  PTRACE(3, "H225\tFast start accepted by remote endpoint");

  PINDEX i;

  // Go through provided list of structures, if can decode it and match it up
  // with a channel we requested AND it has all the information needed in the
  // m_multiplexParameters, then we can start the channel.
  for (i = 0; i < array.GetSize(); i++) {
    H245_OpenLogicalChannel open;
    PPER_Stream strm = (PBYTEArray)array[i];
    if (open.Decode(strm)) {
      PTRACE(4, "H225\tFast start open:\n  " << setprecision(2) << open);
      BOOL reverse = open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters);
      H323Capability * capability;
      if (reverse)
        capability = endpoint.FindCapability(open.m_reverseLogicalChannelParameters.m_dataType);
      else
        capability = endpoint.FindCapability(open.m_forwardLogicalChannelParameters.m_dataType);
      if (capability != NULL) {
        for (PINDEX ch = 0; ch < fastStartChannels.GetSize(); ch++) {
          H323Channel & chan = fastStartChannels[ch];
          if (chan.GetCapability() == *capability &&
             (chan.GetDirection() == H323Channel::IsReceiver) == reverse) {
            unsigned error;
            if (chan.OnReceivedPDU(open, error))
              chan.Start();
            else
              PTRACE(2, "H225\tFast start channel open error: " << error);
            break;
          }
        }
      }
    }
  }

  // Remove any channels that were not started by above, those that were
  // started are put into the logical channel dictionary
  for (i = 0; i < fastStartChannels.GetSize(); i++) {
    if (fastStartChannels[i].IsRunning())
      logicalChannels->Add(fastStartChannels[i]);
    else
      fastStartChannels.RemoveAt(i--);
  }

  // The channels we just transferred to the logical channels dictionary
  // should not be deleted via this structure now.
  fastStartChannels.DisallowDeleteObjects();

  PTRACE(2, "H225\tFast starting " << fastStartChannels.GetSize() << " channels");
  if (fastStartChannels.IsEmpty())
    return FALSE;

  endpoint.OnConnectionEstablished(*this, callToken);
  return TRUE;
}


BOOL H323Connection::StartControlChannel(const H225_TransportAddress & h245Address)
{
  // Check that it is an IP address, all we support at the moment
  if (h245Address.GetTag() != H225_TransportAddress::e_ipAddress) {
    PTRACE(1, "H225\tConnect of H245 failed: Unsupported transport");
    return FALSE;
  }

  H323TransportTCP * channel = new H323TransportTCP(endpoint);
  if (channel->SetRemoteAddress(h245Address) && channel->ConnectControlChannel(*this)) {
    controlChannel = channel;
    return TRUE;
  }

  PTRACE(1, "H225\tConnect of H245 failed: " << channel->GetErrorText());
  delete channel;
  return FALSE;
}


BOOL H323Connection::OnUnknownSignalPDU(const H323SignalPDU & pdu)
{
  PTRACE(2, "H225\tUnknown signalling PDU: " << pdu);
  return TRUE;
}


BOOL H323Connection::WriteControlPDU(const H323ControlPDU & pdu)
{
  PPER_Stream strm;
  pdu.Encode(strm);
  strm.CompleteEncoding();

  PTRACE(4, "H245\tSending PDU:\n  " << setprecision(2) << pdu
                          << "\n " << setprecision(2) << strm);

  if (!h245Tunneling) {
    if (controlChannel != NULL && controlChannel->WritePDU(strm))
      return TRUE;

    PTRACE(1, "H245\tWrite fail.");
    return FALSE;
  }

  // If have a pending signalling PDU, use it rather than separate write
  H323SignalPDU localTunnelPDU;
  H323SignalPDU * tunnelPDU;
  if (h245TunnelPDU != NULL)
    tunnelPDU = h245TunnelPDU;
  else {
    localTunnelPDU.BuildFacility(callReference, FALSE, callIdentifier, TRUE);
    tunnelPDU = &localTunnelPDU;
  }

  tunnelPDU->m_h323_uu_pdu.IncludeOptionalField(H225_H323_UU_PDU::e_h245Control);
  PINDEX last = tunnelPDU->m_h323_uu_pdu.m_h245Control.GetSize();
  tunnelPDU->m_h323_uu_pdu.m_h245Control.SetSize(last+1);
  tunnelPDU->m_h323_uu_pdu.m_h245Control[last] = strm;

  if (h245TunnelPDU != NULL)
    return TRUE;

  return WriteSignalPDU(localTunnelPDU);
}


BOOL H323Connection::StartControlNegotiations()
{
  PTRACE(2, "H245\tStarted control channel");

  // Begin the capability exchange procedure
  if (!capabilityExchangeProcedure->HasSentCapabilities() &&
      !capabilityExchangeProcedure->Start()) {
    PTRACE(1, "H245\tStart of Capability Exchange failed");
    return FALSE;
  }

  // Begin the Master/Slave determination procedure
  if (!masterSlaveDeterminationProcedure->IsDetermined() &&
      !masterSlaveDeterminationProcedure->Start()) {
    PTRACE(1, "H245\tStart of Master/Slave determination failed");
    return FALSE;
  }

  return TRUE;
}


void H323Connection::HandleControlChannel()
{
  if (!StartControlNegotiations())
    return;

  BOOL ok = TRUE;
  while (ok) {
    PPER_Stream strm;
    if (controlChannel->ReadPDU(strm)) {
      PTRACE(4, "H245\tReceived TPKT: " << strm);
      ok = HandleControlData(strm);
    }
    else {
      PTRACE(1, "H245\tRead error: " << controlChannel->GetErrorText());
      ok = FALSE;
    }
  }

  controlChannel->CloseControlChannel();
  PTRACE(2, "H245\tControl channel closed.");

  Close();
}


BOOL H323Connection::HandleControlData(PPER_Stream & strm)
{
  while (!strm.IsAtEnd()) {
    H323ControlPDU pdu;
    if (!pdu.Decode(strm)) {
      PTRACE(1, "H245\tInvalid PDU read.");
      return TRUE;
    }

    if (!HandleControlPDU(pdu))
      return FALSE;

    if (connectionState == HasExecutedSignalConnect &&
        masterSlaveDeterminationProcedure->IsDetermined() &&
        capabilityExchangeProcedure->HasSentCapabilities() &&
        capabilityExchangeProcedure->HasReceivedCapabilities()) {
      connectionState = EstablishedConnection;
      if (!OnControlChannelOpen())
        return FALSE;
    }

    strm.ByteAlign();
  }

  return TRUE;
}


BOOL H323Connection::HandleControlPDU(const H323ControlPDU & pdu)
{
  PTRACE(4, "H245\tReceived PDU:\n  " << setprecision(2) << pdu);

  switch (pdu.GetTag()) {
    case H245_MultimediaSystemControlMessage::e_request :
      return OnH245Request(pdu);

    case H245_MultimediaSystemControlMessage::e_response :
      return OnH245Response(pdu);

    case H245_MultimediaSystemControlMessage::e_command :
      return OnH245Command(pdu);

    case H245_MultimediaSystemControlMessage::e_indication :
      return OnH245Indication(pdu);
  }

  return OnUnknownControlPDU(pdu);
}


BOOL H323Connection::OnUnknownControlPDU(const H323ControlPDU & pdu)
{
  PTRACE(2, "H245\tUnknown Control PDU: " << pdu);

  H323ControlPDU reply;
  reply.BuildFunctionNotUnderstood(pdu);
  return WriteControlPDU(reply);
}


BOOL H323Connection::OnH245Request(const H323ControlPDU & pdu)
{
  const H245_RequestMessage & request = pdu;

  switch (request.GetTag()) {
    case H245_RequestMessage::e_masterSlaveDetermination :
      return masterSlaveDeterminationProcedure->HandleIncoming(request);

    case H245_RequestMessage::e_terminalCapabilitySet :
      return capabilityExchangeProcedure->HandleIncoming(request);

    case H245_RequestMessage::e_openLogicalChannel :
      return logicalChannels->HandleOpen(request);

    case H245_RequestMessage::e_closeLogicalChannel :
      return logicalChannels->HandleClose(request);

    case H245_RequestMessage::e_requestChannelClose :
      return logicalChannels->HandleRequestClose(request);

    case H245_RequestMessage::e_requestMode :
      return requestModeProcedure->HandleRequest(request);

    case H245_RequestMessage::e_roundTripDelayRequest :
      return roundTripDelayProcedure->HandleRequest(request);
  }

  return OnUnknownControlPDU(pdu);
}


BOOL H323Connection::OnH245Response(const H323ControlPDU & pdu)
{
  const H245_ResponseMessage & response = pdu;

  switch (response.GetTag()) {
    case H245_ResponseMessage::e_masterSlaveDeterminationAck :
      return masterSlaveDeterminationProcedure->HandleAck(response);

    case H245_ResponseMessage::e_masterSlaveDeterminationReject :
      return masterSlaveDeterminationProcedure->HandleReject(response);

    case H245_ResponseMessage::e_terminalCapabilitySetAck :
      return capabilityExchangeProcedure->HandleAck(response);

    case H245_ResponseMessage::e_terminalCapabilitySetReject :
      return capabilityExchangeProcedure->HandleReject(response);

    case H245_ResponseMessage::e_openLogicalChannelAck :
      return logicalChannels->HandleOpenAck(response);

    case H245_ResponseMessage::e_openLogicalChannelReject :
      return logicalChannels->HandleReject(response);

    case H245_ResponseMessage::e_closeLogicalChannelAck :
      return logicalChannels->HandleCloseAck(response);

    case H245_ResponseMessage::e_requestChannelCloseAck :
      return logicalChannels->HandleRequestCloseAck(response);

    case H245_ResponseMessage::e_requestChannelCloseReject :
      return logicalChannels->HandleRequestCloseReject(response);

    case H245_ResponseMessage::e_requestModeAck :
      return requestModeProcedure->HandleAck(response);

    case H245_ResponseMessage::e_requestModeReject :
      return requestModeProcedure->HandleReject(response);

    case H245_ResponseMessage::e_roundTripDelayResponse :
      return roundTripDelayProcedure->HandleResponse(response);
  }

  return OnUnknownControlPDU(pdu);
}


BOOL H323Connection::OnH245Command(const H323ControlPDU & pdu)
{
  const H245_CommandMessage & command = pdu;

  switch (command.GetTag()) {
    case H245_CommandMessage::e_sendTerminalCapabilitySet :
      return OnH245_SendTerminalCapabilitySet(command);

    case H245_CommandMessage::e_flowControlCommand :
      return OnH245_FlowControlCommand(command);

    case H245_CommandMessage::e_miscellaneousCommand :
      return OnH245_MiscellaneousCommand(command);

    case H245_CommandMessage::e_endSessionCommand :
      return FALSE;
  }

  return OnUnknownControlPDU(pdu);
}


BOOL H323Connection::OnH245Indication(const H323ControlPDU & pdu)
{
  const H245_IndicationMessage & indication = pdu;

  switch (indication.GetTag()) {
    case H245_IndicationMessage::e_masterSlaveDeterminationRelease :
      return masterSlaveDeterminationProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_terminalCapabilitySetRelease :
      return capabilityExchangeProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_openLogicalChannelConfirm :
      return logicalChannels->HandleOpenConfirm(indication);

    case H245_IndicationMessage::e_requestChannelCloseRelease :
      return logicalChannels->HandleRequestCloseRelease(indication);

    case H245_IndicationMessage::e_requestModeRelease :
      return requestModeProcedure->HandleRelease(indication);

    case H245_IndicationMessage::e_miscellaneousIndication :
      OnH245_MiscellaneousIndication(indication);

    case H245_IndicationMessage::e_jitterIndication :
      OnH245_JitterIndication(indication);

    case H245_IndicationMessage::e_userInput :
      OnUserInputIndication(indication);
  }

  return TRUE; // Do NOT call OnUnknownControlPDU for indications
}


BOOL H323Connection::OnH245_SendTerminalCapabilitySet(
                 const H245_SendTerminalCapabilitySet & pdu)
{
  if (pdu.GetTag() == H245_SendTerminalCapabilitySet::e_genericRequest)
    return capabilityExchangeProcedure->Start();

  PTRACE(2, "H245\tUnhandled SendTerminalCapabilitySet: " << pdu);
  return TRUE;
}


BOOL H323Connection::OnH245_FlowControlCommand(
                 const H245_FlowControlCommand & pdu)
{
  PTRACE(3, "H245\tFlowControlCommand: scope=" << pdu.m_scope.GetTagName());

  long restriction;
  if (pdu.m_restriction.GetTag() == H245_FlowControlCommand_restriction::e_maximumBitRate)
    restriction = (const PASN_Integer &)pdu.m_restriction;
  else
    restriction = -1; // H245_FlowControlCommand_restriction::e_noRestriction

  switch (pdu.m_scope.GetTag()) {
    case H245_FlowControlCommand_scope::e_wholeMultiplex :
      OnLogicalChannelFlowControl(NULL, restriction);
      break;

    case H245_FlowControlCommand_scope::e_logicalChannelNumber :
    {
      H245NegLogicalChannel * negChan = logicalChannels->FindChannel((unsigned)(const H245_LogicalChannelNumber &)pdu.m_scope, FALSE);
      if (negChan != NULL && negChan->GetChannel() != NULL)
        OnLogicalChannelFlowControl(negChan->GetChannel(), restriction);
    }
  }

  return TRUE;
}


BOOL H323Connection::OnH245_MiscellaneousCommand(
                 const H245_MiscellaneousCommand & pdu)
{
  PTRACE(3, "H245\tMiscellaneousCommand: " << pdu.m_logicalChannelNumber);

  H245NegLogicalChannel * negChan = logicalChannels->FindChannel((unsigned)pdu.m_logicalChannelNumber, FALSE);
  if (negChan != NULL && negChan->GetChannel() != NULL)
    negChan->GetChannel()->OnMiscellaneousCommand(pdu.m_type);

  return TRUE;
}


BOOL H323Connection::OnH245_MiscellaneousIndication(
                 const H245_MiscellaneousIndication & pdu)
{
  PTRACE(3, "H245\tMiscellaneousIndication: " << pdu.m_logicalChannelNumber);

  H245NegLogicalChannel * negChan = logicalChannels->FindChannel((unsigned)pdu.m_logicalChannelNumber, FALSE);
  if (negChan != NULL && negChan->GetChannel() != NULL)
    negChan->GetChannel()->OnMiscellaneousIndication(pdu.m_type);

  return TRUE;
}


BOOL H323Connection::OnH245_JitterIndication(
                 const H245_JitterIndication & pdu)
{
  PTRACE(3, "H245\tJitterIndication: scope=" << pdu.m_scope.GetTagName());

  static const DWORD mantissas[8] = { 0, 1, 10, 100, 1000, 10000, 100000, 1000000 };
  static const DWORD exponents[8] = { 10, 25, 50, 75 };
  DWORD jitter = mantissas[pdu.m_estimatedReceivedJitterMantissa]*
                 exponents[pdu.m_estimatedReceivedJitterExponent]/10;

  int skippedFrameCount = -1;
  if (pdu.HasOptionalField(H245_JitterIndication::e_skippedFrameCount))
    skippedFrameCount = pdu.m_skippedFrameCount;

  int additionalBuffer = -1;
  if (pdu.HasOptionalField(H245_JitterIndication::e_additionalDecoderBuffer))
    additionalBuffer = pdu.m_additionalDecoderBuffer;

  switch (pdu.m_scope.GetTag()) {
    case H245_JitterIndication_scope::e_wholeMultiplex :
      OnLogicalChannelJitter(NULL, jitter, skippedFrameCount, additionalBuffer);
      break;

    case H245_JitterIndication_scope::e_logicalChannelNumber :
    {
      H245NegLogicalChannel * negChan = logicalChannels->FindChannel((unsigned)(const H245_LogicalChannelNumber &)pdu.m_scope, FALSE);
      if (negChan != NULL && negChan->GetChannel() != NULL)
        OnLogicalChannelJitter(negChan->GetChannel(), jitter, skippedFrameCount, additionalBuffer);
    }
  }

  return TRUE;
}


BOOL H323Connection::OnControlProtocolError(ControlProtocolErrors /*errorSource*/,
                                            const void * /*errorData*/)
{
  return TRUE;
}


void H323Connection::OnSendCapabilitySet(H323ControlPDU & /*pdu*/)
{
}


BOOL H323Connection::OnReceivedCapabilitySet(const H323CapabilityList & remoteTable,
                                             const H323CapabilitySet & remoteSet,
                                             const H245_MultiplexCapability * muxCap,
                                             H323ControlPDU & /*rejectPDU*/)
{
  // This should actually merge the tables rather than just replace it!
  remoteCapabilityTable = remoteTable;
  remoteCapabilitySet = remoteSet;

  if (muxCap != NULL) {
    if (muxCap->GetTag() != H245_MultiplexCapability::e_h2250Capability) {
      PTRACE(1, "H323\tCapabilitySet contains unsupported multiplex.");
      return FALSE;
    }

    const H245_H2250Capability & h225_0 = *muxCap;
    remoteMaxAudioDelayJitter = h225_0.m_maximumAudioDelayJitter;
  }


  PTRACE(3, "H245\tReceived capability set, is "
                 << (remoteTable.IsEmpty() ? "rejected" : "accepted"));
  return !remoteTable.IsEmpty();
}


BOOL H323Connection::IsH245Master() const
{
  return masterSlaveDeterminationProcedure->IsMaster();
}


BOOL H323Connection::OnControlChannelOpen()
{
  endpoint.OnConnectionEstablished(*this, callToken);
  OnSelectLogicalChannels();
  return TRUE;
}


void H323Connection::OnSelectLogicalChannels()
{
  PTRACE(2, "H245\tDefault OnSelectLogicalChannels, " << fastStartState);

  PINDEX i;
  const H323CapabilityList & localCapabilityTable = endpoint.GetCapabilityTable();

  // Select the first codec that uses the "standard" audio session.
  switch (fastStartState) {
    case FastStartDisabled :
      for (i = 0; i < localCapabilityTable.GetSize(); i++) {
        H323Capability & capability = localCapabilityTable[i];
        if (capability.GetDefaultSessionID() == RTP_Session::DefaultAudioSessionID) {
          PINDEX idx = remoteCapabilityTable.GetValuesIndex(capability);
          if (idx != P_MAX_INDEX) {
            OpenLogicalChannel(remoteCapabilityTable[idx],
                               RTP_Session::DefaultAudioSessionID,
                               H323Channel::IsTransmitter);
            break;
          }
        }
      }
      break;

    case FastStartInitiate :
      for (i = 0; i < localCapabilityTable.GetSize(); i++) {
        H323Capability & capability = localCapabilityTable[i];
        if (capability.GetDefaultSessionID() == RTP_Session::DefaultAudioSessionID) {
          OpenLogicalChannel(capability,
                             RTP_Session::DefaultAudioSessionID,
                             H323Channel::IsReceiver);
          OpenLogicalChannel(capability,
                             RTP_Session::DefaultAudioSessionID,
                             H323Channel::IsTransmitter);
          break;
        }
      }
      break;

    default : // FastStartResponse
      for (i = 0; i < fastStartChannels.GetSize(); i++) {
        H323Channel & channel = fastStartChannels[i];
        if (channel.GetSessionID() == RTP_Session::DefaultAudioSessionID &&
            channel.GetDirection() == H323Channel::IsTransmitter) {
          fastStartChannels[i].Start();
          break;
        }
      }
      for (i = 0; i < fastStartChannels.GetSize(); i++) {
        H323Channel & channel = fastStartChannels[i];
        if (channel.GetSessionID() == RTP_Session::DefaultAudioSessionID &&
            channel.GetDirection() == H323Channel::IsReceiver) {
          fastStartChannels[i].Start();
          break;
        }
      }
      break;
  }
}


BOOL H323Connection::OpenLogicalChannel(const H323Capability & capability,
                                        unsigned sessionID,
                                        H323Channel::Directions dir)
{
  switch (fastStartState) {
    case FastStartDisabled :
      if (dir == H323Channel::IsReceiver)
        return FALSE;

      // Traditional H245 handshake
      return logicalChannels->Open(capability, sessionID);

    case FastStartResponse :
      // Do not use OpenLogicalChannel for starting these.
      return FALSE;

    default : // FastStartInitiate
      break;
  }

  /*If starting a receiver channel and are initiating the fast start call,
    indicated by the remoteCapabilityTable being empty, we do a "trial"
    listen on the channel. That is, for example, the UDP sockets are created
    to receive data in the RTP session, but no thread is started to read the
    packets and pass them to the codec. This is because at this point in time,
    we do not know which of the codecs is to be used, and more than one thread
    cannot read from the RTP ports at the same time.
  */
  H323Channel * channel = capability.CreateChannel(*this, dir, sessionID, NULL);
  if (channel == NULL)
    return FALSE;

  if (!channel->SetInitialBandwidth()) {
    delete channel;
    return FALSE;
  }

  if (dir != H323Channel::IsReceiver)
    channel->SetNumber(logicalChannels->GetNextChannelNumber());

  fastStartChannels.Append(channel);
  return TRUE;
}


int H323Connection::OnOpenLogicalChannel(const H245_OpenLogicalChannel & /*openPDU*/,
                                         H245_OpenLogicalChannelAck & /*ackPDU*/,
                                         unsigned & /*errorCode*/)
{
  //errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
  return TRUE;
}


BOOL H323Connection::OnStartLogicalChannel(H323Channel & channel)
{
  return endpoint.OnStartLogicalChannel(*this, channel);
}


void H323Connection::CloseLogicalChannel(unsigned number, BOOL fromRemote)
{
  logicalChannels->Close(number, fromRemote);
}


void H323Connection::OnClosedLogicalChannel(const H323Channel & /*channel*/)
{
}


BOOL H323Connection::OnClosingLogicalChannel(H323Channel & /*channel*/)
{
  return TRUE;
}


void H323Connection::OnLogicalChannelFlowControl(H323Channel * channel,
                                                 long bitRateRestriction)
{
  if (channel != NULL)
    channel->OnFlowControl(bitRateRestriction);
}


void H323Connection::OnLogicalChannelJitter(H323Channel * channel,
                                            DWORD jitter,
                                            int skippedFrameCount,
                                            int additionalBuffer)
{
  if (channel != NULL)
    channel->OnJitterIndication(jitter, skippedFrameCount, additionalBuffer);
}


unsigned H323Connection::GetBandwidthUsed() const
{
  unsigned used = 0;

  for (PINDEX i = 0; i < logicalChannels->GetSize(); i++) {
    H323Channel * channel = logicalChannels->GetChannelAt(i);
    if (channel != NULL)
      used += channel->GetBandwidthUsed();
  }

  PTRACE(3, "H323\tBandwidth used: " << used);

  return used;
}


BOOL H323Connection::UseBandwidth(unsigned bandwidth, BOOL removing)
{
  PTRACE(3, "H323\tBandwidth request: " << (removing ? '+' : '-') << bandwidth);

  if (removing)
    bandwidthAvailable += bandwidth;
  else {
    if (bandwidth > bandwidthAvailable) {
      PTRACE(2, "H323\tAvailable bandwidth exceeded");
      return FALSE;
    }

    bandwidthAvailable -= bandwidth;
  }

  return TRUE;
}


BOOL H323Connection::SetBandwidthAvailable(unsigned newBandwidth, BOOL force)
{
  unsigned used = GetBandwidthUsed();
  if (used > newBandwidth) {
    if (!force)
      return FALSE;

    // Go through logical channels and close down some.
    PINDEX chanIdx = logicalChannels->GetSize();
    while (used > newBandwidth && chanIdx-- > 0) {
      H323Channel * channel = logicalChannels->GetChannelAt(chanIdx);
      if (channel != NULL) {
        used -= channel->GetBandwidthUsed();
        const H323ChannelNumber & number = channel->GetNumber();
        CloseLogicalChannel(number, number.IsFromRemote());
      }
    }
  }

  bandwidthAvailable = newBandwidth - used;
  return TRUE;
}


void H323Connection::SendUserInput(const H245_UserInputIndication & indication)
{
  H323ControlPDU pdu;
  H245_UserInputIndication & ind = pdu.Build(H245_IndicationMessage::e_userInput);
  ind = indication;
  WriteControlPDU(pdu);
}


void H323Connection::SendUserInput(const PString & value)
{
  H323ControlPDU pdu;
  pdu.BuildUserInputIndication(value);
  WriteControlPDU(pdu);
}


void H323Connection::OnUserInputIndication(const H245_UserInputIndication & ind)
{
  if (ind.GetTag() == H245_UserInputIndication::e_alphanumeric)
    OnUserInputString((const PASN_GeneralString &)ind);
}


void H323Connection::OnUserInputString(const PString &)
{
}


RTP_Session * H323Connection::GetSession(unsigned sessionID) const
{
  PWaitAndSignal mutex(((H323Connection*)this)->rtpSessionMutex);
  if (!rtpSessions.Contains(sessionID))
    return FALSE;

  PTRACE(3, "RTP\tFound existing session " << sessionID);
  return &rtpSessions[sessionID];
}


PString H323Connection::GetSessionCodecNames(unsigned sessionID) const
{
  PStringStream name;

  for (PINDEX i = 0; i < logicalChannels->GetSize(); i++) {
    H323Channel * channel = logicalChannels->GetChannelAt(i);
    if (channel != NULL && channel->GetSessionID() == sessionID) {
      H323Codec * codec = channel->GetCodec();
      if (codec != NULL) {
        if (name.IsEmpty())
          name << *codec;
        else
          name << " / " << *codec;
      }
    }
  }

  return name;
}


RTP_Session * H323Connection::UseSession(unsigned sessionID,
                                         const H245_TransportAddress & taddr)
{
  PWaitAndSignal mutex(rtpSessionMutex);

  if (rtpSessions.Contains(sessionID)) {
    PTRACE(3, "RTP\tFound existing session " << sessionID);
    rtpSessions[sessionID].IncrementReference();
    return &rtpSessions[sessionID];
  }

  // We only support unicast IP at this time.
  if (taddr.GetTag() != H245_TransportAddress::e_unicastAddress)
    return NULL;

  const H245_UnicastAddress & uaddr = taddr;
  if (uaddr.GetTag() != H245_UnicastAddress::e_iPAddress)
    return NULL;

//  H245_UnicastAddress_iPAddress & ipaddr = uaddr;
//  PIPSocket::Address ipnum(ipaddr.m_network[0], ipaddr.m_network[1],
//                           ipaddr.m_network[2], ipaddr.m_network[3]);

  PTRACE(2, "RTP\tCreating session " << sessionID);
  RTP_Session * session = new RTP_UDP(*this, sessionID, endpoint.GetRtpPortBase());
  rtpSessions.SetAt(sessionID, session);
  return session;
}


void H323Connection::ReleaseSession(unsigned sessionID)
{
  PTRACE(2, "RTP\tReleasing session " << sessionID);

  rtpSessionMutex.Wait();

  if (rtpSessions.Contains(POrdinalKey(sessionID))) {
    if (rtpSessions[sessionID].DecrementReference()) {
      PTRACE(3, "RTP\tDeleting session " << sessionID);
      rtpSessions.SetAt(sessionID, NULL);
    }
  }

  rtpSessionMutex.Signal();
}


void H323Connection::RequestModeChange()
{
}


BOOL H323Connection::OnRequestModeChange(const H245_RequestMode & /*pdu*/,
                                         H245_RequestModeAck & /*ack*/,
                                         H245_RequestModeReject & /*reject*/)
{
  return TRUE;
}


void H323Connection::OnAcceptModeChange(const H245_RequestModeAck & /*pdu*/)
{
}


void H323Connection::OnRefusedModeChange(const H245_RequestModeReject * /*pdu*/)
{
}


const H323Transport & H323Connection::GetControlChannel() const
{
  return *(controlChannel != NULL ? controlChannel : signallingChannel);
}


/////////////////////////////////////////////////////////////////////////////
