/*
 * h323ep.cxx
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: h323ep.cxx,v $
 * Revision 1.4  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.3  1999/08/27 15:42:44  craigs
 * Fixed problem with local call tokens using ambiguous interface names, and connect timeouts not changing connection state
 *
 * Revision 1.2  1999/08/27 09:46:05  robertj
 * Added sepearte function to initialise vendor information from endpoint.
 *
 * Revision 1.1  1999/08/25 05:10:36  robertj
 * File fission (critical mass reached).
 * Improved way in which remote capabilities are created, removed case statement!
 * Changed MakeCall, so immediately spawns thread, no black on TCP connect.
 *
 */

#include <ptlib.h>
#include "h323.h"
#include "h323pdu.h"


class H225CallThread : public PThread
{
  PCLASSINFO(H225CallThread, PThread)

  public:
    H225CallThread(H323EndPoint & endpoint,
                   H323Connection & connection,
                   H323Transport & transport,
                   const PString & remoteParty);

  protected:
    void Main();

    H323Connection & connection;
    H323Transport  & transport;
    PString          remoteParty;
};


class H323ConnectionsCleaner : public PThread
{
  PCLASSINFO(H323ConnectionsCleaner, PThread)

  public:
    H323ConnectionsCleaner(H323EndPoint & endpoint);

  protected:
    void Main();

    H323EndPoint & endpoint;
};


#define new PNEW


/////////////////////////////////////////////////////////////////////////////

H225CallThread::H225CallThread(H323EndPoint & endpoint,
                               H323Connection & c,
                               H323Transport & t,
                               const PString & r)
  : PThread(endpoint.GetSignallingThreadStackSize(), NoAutoDeleteThread),
    connection(c),
    transport(t),
    remoteParty(r)
{
  transport.AttachThread(this);
  Resume();
}


void H225CallThread::Main()
{
  PTRACE(3, "H225\tStarted call thread");

  if (connection.SendSignalSetup(remoteParty))
    connection.HandleSignallingChannel();
  else
    connection.GetEndPoint().SignalCleanUpConnections();
}


/////////////////////////////////////////////////////////////////////////////

H323ConnectionsCleaner::H323ConnectionsCleaner(H323EndPoint & end)
  : PThread(10000, NoAutoDeleteThread),
    endpoint(end)
{
  Resume();
}


void H323ConnectionsCleaner::Main()
{
  endpoint.CleanUpConnections();
}


/////////////////////////////////////////////////////////////////////////////

H323EndPoint::H323EndPoint()
  : localUserName(PProcess::Current().GetUserName()),
    soundChannelPlayDevice(PSoundChannel::GetDefaultDevice(PSoundChannel::Player)),
    soundChannelRecordDevice(PSoundChannel::GetDefaultDevice(PSoundChannel::Recorder)),
    signallingChannelCallTimeout(0, 60),    // Seconds
    masterSlaveDeterminationTimeout(0, 30), // Seconds
    capabilityExchangeTimeout(0, 30),       // Seconds
    logicalChannelTimeout(0, 30),           // Seconds
    requestModeTimeout(0, 30),              // Seconds
    roundTripDelayTimeout(0, 30),           // Seconds
    gatekeeperRequestTimeout(0, 5),         // Seconds
    rasRequestTimeout(0, 5)                 // Seconds
{
  terminalType = e_TerminalOnly;
  maxAudioDelayJitter = 150; // milliseconds
  initialBandwidth = 1280;   // 2 x 64k (G.711) channels in 100's of bits/sec

  t35CountryCode = 181;
  t35Extension = 0;
  manufacturerCode = 31415;

  rtpPortBase = 5000;

  masterSlaveDeterminationRetries = 10;
  gatekeeperRequestRetries = 2;
  rasRequestRetries = 2;

  signallingThreadStackSize = 30000;
  controlThreadStackSize    = 30000;
  logicalThreadStackSize    = 30000;
  listenerThreadStackSize   = 30000;
  rasThreadStackSize        = 30000;

  gatekeeper = NULL;
  gatekeeperDiscoverySequenceNumber = 0;

  connectionsCleanerStop = FALSE;
  connectionsCleaner = new H323ConnectionsCleaner(*this);

  srand((unsigned)time(NULL)+clock());

  PTRACE(3, "H323\tCreated endpoint.");
}


H323EndPoint::~H323EndPoint()
{
  // Shut down the listeners as soon as possible to avoid race conditions
  listeners.RemoveAll();

  // Shut down the cleaner thread
  connectionsCleanerStop = TRUE;
  connectionsSignal.Signal();
  connectionsCleaner->WaitForTermination();
  delete connectionsCleaner;

  // And shut down the gatekeeper (if there was one)
  RemoveGatekeeper();

  ClearAllCalls();

  // And clean up the objects
  connections.RemoveAll();
}


void H323EndPoint::SetEndpointTypeInfo(H225_EndpointType & info) const
{
  info.IncludeOptionalField(H225_EndpointType::e_vendor);
  SetVendorIdentifierInfo(info.m_vendor);

  switch (terminalType) {
    case e_TerminalOnly :
    case e_TerminalAndMC :
      info.IncludeOptionalField(H225_EndpointType::e_terminal);
      break;
    case e_GatewayOnly :
    case e_GatewayAndMC :
    case e_GatewayAndMCWithDataMP :
    case e_GatewayAndMCWithAudioMP :
    case e_GatewayAndMCWithAVMP :
      info.IncludeOptionalField(H225_EndpointType::e_gateway);
      break;
    case e_GatekeeperOnly :
    case e_GatekeeperWithDataMP :
    case e_GatekeeperWithAudioMP :
    case e_GatekeeperWithAVMP :
      info.IncludeOptionalField(H225_EndpointType::e_gatekeeper);
      break;
    case e_MCUOnly :
    case e_MCUWithDataMP :
    case e_MCUWithAudioMP :
    case e_MCUWithAVMP :
      info.IncludeOptionalField(H225_EndpointType::e_mcu);
      info.m_mc = TRUE;
  }
}


void H323EndPoint::SetVendorIdentifierInfo(H225_VendorIdentifier & info) const
{
  SetH221NonStandardInfo(info.m_vendor);

  info.IncludeOptionalField(H225_VendorIdentifier::e_productId);
  info.m_productId = PProcess::Current().GetManufacturer() & PProcess::Current().GetName();
  info.m_productId.SetSize(info.m_productId.GetSize()+2);

  info.IncludeOptionalField(H225_VendorIdentifier::e_versionId);
  info.m_versionId = PProcess::Current().GetVersion(TRUE);
  info.m_versionId.SetSize(info.m_versionId.GetSize()+2);
}


void H323EndPoint::SetH221NonStandardInfo(H225_H221NonStandard & info) const
{
  info.m_t35CountryCode = t35CountryCode;
  info.m_t35Extension = t35Extension;
  info.m_manufacturerCode = manufacturerCode;
}


H323Capability * H323EndPoint::FindCapability(const H245_Capability & cap)
{
  PTRACE(4, "H323\tFindCapability: " << cap.GetTagName());

  switch (cap.GetTag()) {
    case H245_Capability::e_receiveAudioCapability :
    case H245_Capability::e_transmitAudioCapability :
    case H245_Capability::e_receiveAndTransmitAudioCapability :
    {
      const H245_AudioCapability & audio = cap;
      return FindCapability(H323Capability::e_Audio, audio.GetTag());
    }

    case H245_Capability::e_receiveVideoCapability :
    case H245_Capability::e_transmitVideoCapability :
    case H245_Capability::e_receiveAndTransmitVideoCapability :
    {
      const H245_VideoCapability & video = cap;
      return FindCapability(H323Capability::e_Video, video.GetTag());
    }

    case H245_Capability::e_receiveDataApplicationCapability :
    case H245_Capability::e_transmitDataApplicationCapability :
    case H245_Capability::e_receiveAndTransmitDataApplicationCapability :
    {
      const H245_DataApplicationCapability & data = cap;
      return FindCapability(H323Capability::e_Data, data.m_application.GetTag());
    }

    case H245_Capability::e_receiveUserInputCapability :
    case H245_Capability::e_transmitUserInputCapability :
    case H245_Capability::e_receiveAndTransmitUserInputCapability :
    {
      const H245_UserInputCapability & ui = cap;
      return FindCapability(H323Capability::e_UserInput, ui.GetTag());
    }

    default :
      break;
  }

  return NULL;
}


H323Capability * H323EndPoint::FindCapability(const H245_DataType & dataType)
{
  PTRACE(4, "H323\tFindCapability: " << dataType.GetTagName());

  switch (dataType.GetTag()) {
    case H245_DataType::e_audioData :
      {
        const H245_AudioCapability & audio = dataType;
        return FindCapability(H323Capability::e_Audio, audio.GetTag());
      }

    case H245_DataType::e_videoData :
      {
        const H245_VideoCapability & video = dataType;
        return FindCapability(H323Capability::e_Video, video.GetTag());
      }

    case H245_DataType::e_data :
      {
        const H245_DataApplicationCapability & data = dataType;
        return FindCapability(H323Capability::e_Data, data.GetTag());
      }

    default :
      break;
  }

  return NULL;
}


H323Capability * H323EndPoint::FindCapability(H323Capability::MainTypes mainType,
                                              unsigned subType)
{
  PTRACE(3, "H323\tFindCapability: " << mainType << " subtype=" << subType);

  for (PINDEX i = 0; i < capabilityTable.GetSize(); i++) {
    if (capabilityTable[i].GetMainType() == mainType &&
        capabilityTable[i].GetSubType() == subType) {
      PTRACE(3, "H323\tFound capability: " << capabilityTable[i]);
      return &capabilityTable[i];
    }
  }

  return NULL;
}


void H323EndPoint::AddCapability(H323Capability * capability)
{
  // See if already added, confuses things if you add the same instance twice
  if (capabilityTable.GetObjectsIndex(capability) != P_MAX_INDEX)
    return;

  // Assign a unique number to the codec, check if the user wants a specific
  // value and start with that.
  capability->AssignCapabilityCode(capabilityTable);
  capabilityTable.Append(capability);
  PTRACE(2, "H323\tAdded capability: " << *capability);
}


void H323EndPoint::SetCapability(PINDEX descriptorNum,
                                 PINDEX simultaneousNum,
                                 H323Capability * capability)
{
  // Make sure the outer array is big enough
  capabilitySet.SetMinSize(descriptorNum+1);

  // Make sure the middle array is big enough
  capabilitySet[descriptorNum].SetMinSize(simultaneousNum+1);

  // Now we can put the new entry in.
  capabilitySet[descriptorNum][simultaneousNum].Append(capability);
}


BOOL H323EndPoint::SetGatekeeper(const PString & address, H323Transport * transport)
{
  if (!RemoveGatekeeper(H225_UnregRequestReason::e_reregistrationRequired))
    return FALSE;

  H323Gatekeeper * gk = CreateGatekeeper(transport);
  if (gk->RegistrationRequest(address)) {
    gatekeeper = gk;
    return TRUE;
  }

  delete gk;
  return FALSE;
}


BOOL H323EndPoint::DiscoverGatekeeper(H323Transport * transport)
{
  if (!RemoveGatekeeper(H225_UnregRequestReason::e_reregistrationRequired))
    return FALSE;

  H323Gatekeeper * gk = CreateGatekeeper(transport);
  if (gk->Discover() && gk->RegistrationRequest(PString())) {
    gatekeeper = gk;
    return TRUE;
  }

  delete gk;
  return FALSE;
}


H323Gatekeeper * H323EndPoint::CreateGatekeeper(H323Transport * transport)
{
  return new H323Gatekeeper(*this, transport);
}


BOOL H323EndPoint::RemoveGatekeeper(int reason)
{
  if (gatekeeper == NULL)
    return TRUE;

  ClearAllCalls();

  if (!gatekeeper->UnregistrationRequest(reason))
    return FALSE;

  delete gatekeeper;
  gatekeeper = NULL;
  return TRUE;
}


BOOL H323EndPoint::StartListener(H323Listener * listener)
{
  if (!listener->Open())
    return FALSE;

  listeners.Append(listener);
  listener->Resume();
  return TRUE;
}


BOOL H323EndPoint::MakeCall(const PString & remoteParty,
                            H323Transport * transport,
                            PString & token)
{
  PTRACE(3, "H323\tMaking call to: " << remoteParty);

  static unsigned lastReference = rand()&0x3fff;

  connectionsMutex.Wait();

  do {
    lastReference = (lastReference+1)&0x3fff;
    token = BuildConnectionToken(*transport, lastReference, FALSE);
  } while (connections.Contains(token));

  H323Connection * connection = CreateConnection(lastReference);
  connections.SetAt(token, connection);

  connectionsMutex.Signal();

  PTRACE(3, "H323\tCreated new connection: " << token);

  connection->AttachSignalChannel(transport);
  new H225CallThread(*this, *connection, *transport, remoteParty);
  return TRUE;
}


BOOL H323EndPoint::ClearCall(const PString & token, BOOL gkForced)
{
  H323Connection * connection = FindConnectionWithLock(token);
  if (connection == NULL)
    return FALSE;

  connection->Close(gkForced);
  connection->Unlock();
  return TRUE;
}


void H323EndPoint::ClearAllCalls(BOOL gkForced)
{
  PWaitAndSignal wait(connectionsMutex);
  for (PINDEX i = 0; i < connections.GetSize(); i++)
    connections.GetDataAt(i).Close(gkForced);
}


void H323EndPoint::OnConnectionEstablished(H323Connection & /*connection*/,
                                           const PString & /*token*/)
{
}


void H323EndPoint::OnConnectionCleared(H323Connection & /*connection*/,
                                       const PString & /*token*/)
{
}


H323Connection * H323EndPoint::FindConnectionWithLock(const PString & token)
{
  if (token.IsEmpty())
    return NULL;

  PWaitAndSignal wait(connectionsMutex);

  H323Connection * conn_ptr = connections.GetAt(token);
  if (conn_ptr != NULL){
    conn_ptr->Lock();
    return conn_ptr;
  }

  PINDEX i;
  for (i = 0; i < connections.GetSize(); i++) {
    H323Connection & conn = connections.GetDataAt(i);
    if (conn.GetCallIdentifier().AsString() == token) {
      conn.Lock();
      return &conn;
    }
  }

  for (i = 0; i < connections.GetSize(); i++) {
    H323Connection & conn = connections.GetDataAt(i);
    if (conn.GetConferenceIdentifier().AsString() == token) {
      conn.Lock();
      return &conn;
    }
  }

  return NULL;
}


PString H323EndPoint::BuildConnectionToken(const H323Transport & transport,
                                           unsigned callReference,
                                           BOOL fromRemote)
{
  PString token;

  if (fromRemote)
    token = transport.GetRemoteAddress();
  else
    token = "ip:localhost";

  token.sprintf("/%u", callReference);

  return token;
}


H323Connection & H323EndPoint::GetConnection(const H323Transport & transport,
                                           unsigned callReference,
                                           BOOL fromRemote)
{
  PString token = BuildConnectionToken(transport, callReference, fromRemote);

  PWaitAndSignal wait(connectionsMutex);

  if (connections.Contains(token))
    return connections[token];

  H323Connection * connection = CreateConnection(callReference);
  connections.SetAt(token, connection);

  PTRACE(3, "H323\tCreated new connection: " << token);

  return *connection;
}


void H323EndPoint::CleanUpConnections()
{
  for (;;) {
    connectionsSignal.Wait();
    if (connectionsCleanerStop)
      return;

    connectionsMutex.Wait();

    PINDEX i = 0;
    while (i < connections.GetSize()) {
      H323Connection & connection = connections.GetDataAt(i);
      if (connection.IsOpen())
        i++;
      else {
        connection.Lock();
        PString token = connections.GetKeyAt(i);
        OnConnectionCleared(connection, token);
        connections.SetAt(token, NULL);
        i = 0;
      }
    }

    connectionsMutex.Signal();
  }
}


void H323EndPoint::SignalCleanUpConnections()
{
  connectionsSignal.Signal();
}


H323Channel * H323EndPoint::CreateLogicalChannel(H323Connection & connection,
                                                 const H245_OpenLogicalChannel & open,
                                                 unsigned & errorCode)
{
  const H245_H2250LogicalChannelParameters * param;
  const H245_DataType * dataType;
  H323Channel::Directions direction;

  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
    if (open.m_reverseLogicalChannelParameters.m_multiplexParameters.GetTag() !=
              H245_OpenLogicalChannel_reverseLogicalChannelParameters_multiplexParameters
                                                      ::e_h2250LogicalChannelParameters) {
      errorCode = H245_OpenLogicalChannelReject_cause::e_unsuitableReverseParameters;
      PTRACE(2, "H323\tCreateLogicalChannel - reverse channel, H225.0 only supported");
      return NULL;
    }

    PTRACE(3, "H323\tCreateLogicalChannel - reverse channel");
    dataType = &open.m_reverseLogicalChannelParameters.m_dataType;
    param = &(const H245_H2250LogicalChannelParameters &)
                      open.m_reverseLogicalChannelParameters.m_multiplexParameters;
    direction = H323Channel::IsTransmitter;
  }
  else {
    if (open.m_forwardLogicalChannelParameters.m_multiplexParameters.GetTag() !=
              H245_OpenLogicalChannel_forwardLogicalChannelParameters_multiplexParameters
                                                      ::e_h2250LogicalChannelParameters) {
      PTRACE(2, "H323\tCreateLogicalChannel - forward channel, H225.0 only supported");
      errorCode = H245_OpenLogicalChannelReject_cause::e_unspecified;
      return NULL;
    }

    PTRACE(3, "H323\tCreateLogicalChannel - forward channel");
    dataType = &open.m_forwardLogicalChannelParameters.m_dataType;
    param = &(const H245_H2250LogicalChannelParameters &)
                      open.m_forwardLogicalChannelParameters.m_multiplexParameters;
    direction = H323Channel::IsReceiver;
  }

  // Only support unicast channels so *must* have a media control channel
  if (!param->HasOptionalField(H245_H2250LogicalChannelParameters::e_mediaControlChannel)) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_multicastChannelNotAllowed;
    PTRACE(2, "H323\tCreateLogicalChannel - multicast not supported");
    return NULL;
  }

  // See if datatype is supported
  H323Capability * capability = FindCapability(*dataType);
  if (capability == NULL) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_unknownDataType;
    return NULL; // If codec not supported, return error
  }

  H323Channel * channel = capability->CreateChannel(connection, direction, 0, param);
  if (channel == NULL) {
    errorCode = H245_OpenLogicalChannelReject_cause::e_dataTypeNotAvailable;
    return NULL;
  }

  if (!channel->SetInitialBandwidth())
    errorCode = H245_OpenLogicalChannelReject_cause::e_insufficientBandwidth;
  else if (channel->OnReceivedPDU(open, errorCode))
    return channel;

  delete channel;
  return NULL;
}


BOOL H323EndPoint::OnStartLogicalChannel(H323Connection & connection,
                                         H323Channel & channel)
{
  H323Codec * codec = channel.GetCodec();
  if (codec != NULL)
    return codec->Open(connection);

  return TRUE;
}


PChannel * H323EndPoint::OpenAudioChannel(BOOL isEncoding, unsigned bufferSize)
{
  PString deviceName;
  if (isEncoding)
    deviceName = GetSoundChannelRecordDevice();
  else
    deviceName = GetSoundChannelPlayDevice();

  PSoundChannel * soundChannel = new PSoundChannel;
  soundChannel->SetBuffers(bufferSize, 8);

  if (soundChannel->Open(deviceName,
                         isEncoding ? PSoundChannel::Recorder
                                    : PSoundChannel::Player,
                         1, 8000, 16)) {
    PTRACE(3, "Codec\tOpened sound channel \"" << deviceName
           << "\" for " << (isEncoding ? "record" : "play") << "ing.");
    return soundChannel;
  }

  PTRACE(1, "Codec\tCould not open sound channel \"" << deviceName
         << "\" for " << (isEncoding ? "record" : "play")
         << "ing: " << soundChannel->GetErrorText());

  delete soundChannel;
  return NULL;
}


void H323EndPoint::SetLocalUserName(const PString & name)
{
  localUserName = name;
}


BOOL H323EndPoint::SetSoundChannelPlayDevice(const PString & name)
{
  if (PSoundChannel::GetDeviceNames(PSoundChannel::Player).GetValuesIndex(name) == P_MAX_INDEX)
    return FALSE;

  soundChannelPlayDevice = name;
  return TRUE;
}


BOOL H323EndPoint::SetSoundChannelRecordDevice(const PString & name)
{
  if (PSoundChannel::GetDeviceNames(PSoundChannel::Recorder).GetValuesIndex(name) == P_MAX_INDEX)
    return FALSE;

  soundChannelRecordDevice = name;
  return TRUE;
}


BOOL H323EndPoint::IsTerminal() const
{
  switch (terminalType) {
    case e_TerminalOnly :
    case e_TerminalAndMC :
      return TRUE;

    default :
      return FALSE;
  }
}


BOOL H323EndPoint::IsGateway() const
{
  switch (terminalType) {
    case e_GatewayOnly :
    case e_GatewayAndMC :
    case e_GatewayAndMCWithDataMP :
    case e_GatewayAndMCWithAudioMP :
    case e_GatewayAndMCWithAVMP :
      return TRUE;

    default :
      return FALSE;
  }
}


BOOL H323EndPoint::IsGatekeeper() const
{
  switch (terminalType) {
    case e_GatekeeperOnly :
    case e_GatekeeperWithDataMP :
    case e_GatekeeperWithAudioMP :
    case e_GatekeeperWithAVMP :
      return TRUE;

    default :
      return FALSE;
  }
}


BOOL H323EndPoint::IsMCU() const
{
  switch (terminalType) {
    case e_MCUOnly :
    case e_MCUWithDataMP :
    case e_MCUWithAudioMP :
    case e_MCUWithAVMP :
      return TRUE;

    default :
      return FALSE;
  }
}


/////////////////////////////////////////////////////////////////////////////
