/*
 * q931.cxx
 *
 * Q.931 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: q931.cxx,v $
 * Revision 1.14  1999/08/31 13:54:35  robertj
 * Fixed problem with memory overrun building PDU's
 *
 * Revision 1.13  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.12  1999/08/13 06:34:38  robertj
 * Fixed problem in CallPartyNumber Q.931 encoding.
 * Added field name display to Q.931 protocol.
 *
 * Revision 1.11  1999/08/10 13:14:15  robertj
 * Added Q.931 Called Number field if have "phone number" style destination addres.
 *
 * Revision 1.10  1999/07/16 02:15:30  robertj
 * Fixed more tunneling problems.
 *
 * Revision 1.9  1999/07/09 14:59:59  robertj
 * Fixed GNU C++ compatibility.
 *
 * Revision 1.8  1999/07/09 06:09:50  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.7  1999/06/14 15:19:48  robertj
 * GNU C compatibility
 *
 * Revision 1.6  1999/06/13 12:41:14  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.5  1999/06/09 05:26:20  robertj
 * Major restructuring of classes.
 *
 * Revision 1.4  1999/02/23 11:04:29  robertj
 * Added capability to make outgoing call.
 *
 * Revision 1.3  1999/01/16 01:31:38  robertj
 * Major implementation.
 *
 * Revision 1.2  1999/01/02 04:00:52  robertj
 * Added higher level protocol negotiations.
 *
 * Revision 1.1  1998/12/14 09:13:37  robertj
 * Initial revision
 *
 */

#include <ptlib.h>
#include "q931.h"

#define new PNEW


Q931::Q931()
{
  protocolDiscriminator = 8;  // Q931 always has 00001000
  messageType = NationalEscapeMsg;
  fromDestination = FALSE;
  callReference = 0;
}


Q931 & Q931::operator=(const Q931 & other)
{
  protocolDiscriminator = other.protocolDiscriminator;
  callReference = other.callReference;
  fromDestination = other.fromDestination;
  messageType = other.messageType;

  fields.RemoveAll();
  for (PINDEX i = 0; i < other.fields.GetSize(); i++)
    fields.SetAt(other.fields.GetKeyAt(i), new PBYTEArray(other.fields.GetDataAt(i)));

  return *this;
}


void Q931::BuildFacility(int callRef, BOOL fromDest)
{
  messageType = FacilityMsg;
  callReference = callRef;
  fromDestination = fromDest;
  fields.RemoveAll();
}


void Q931::BuildCallProceeding(int callRef)
{
  messageType = CallProceedingMsg;
  callReference = callRef;
  fromDestination = TRUE;
  fields.RemoveAll();
}


void Q931::BuildAlerting(int callRef)
{
  messageType = AlertingMsg;
  callReference = callRef;
  fromDestination = TRUE;
  fields.RemoveAll();
}


static const BYTE BearerCapabilityData[3] = { 0x88, 0xc0, 0xa5 };

void Q931::BuildSetup(const PString & name)
{
  GenerateCallReference();
  messageType = SetupMsg;
  fromDestination = FALSE;
  fields.RemoveAll();
  SetDisplayName(name);
  SetField(BearerCapabilityField, PBYTEArray(BearerCapabilityData, sizeof(BearerCapabilityData)));
}


void Q931::BuildConnect(int callRef, const PString & name)
{
  messageType = ConnectMsg;
  callReference = callRef;
  fromDestination = TRUE;
  fields.RemoveAll();
  SetDisplayName(name);
  SetField(BearerCapabilityField, PBYTEArray(BearerCapabilityData, sizeof(BearerCapabilityData)));
}


void Q931::BuildReleaseComplete(int callRef, BOOL fromDest)
{
  messageType = ReleaseCompleteMsg;
  callReference = callRef;
  fromDestination = fromDest;
  fields.RemoveAll();
}


BOOL Q931::Decode(const PBYTEArray & data)
{
  // Clear all existing data before reading new
  fields.RemoveAll();

  if (data.GetSize() < 5) // Packet too short
    return FALSE;

  protocolDiscriminator = data[0];

  if (data[1] != 2) // Call reference must be 2 bytes long
    return FALSE;

  callReference = ((data[2]&0x7f) << 8) | data[3];
  fromDestination = (data[2]&0x80) != 0;

  messageType = (MsgTypes)data[4];

  // Have preamble, start getting the fields into buffers
  PINDEX offset = 5;
  while (offset < data.GetSize()) {
    // Get field discriminator
    int discriminator = data[offset++];

    PBYTEArray * item = new PBYTEArray;

    // For discriminator with high bit set there is no data
    if ((discriminator&0x80) == 0) {
      int len = data[offset++];

      if (discriminator == UserUserField) {
        // Special case of User-user field, there is some confusion here as
        // the Q931 documentation claims the length is a single byte,
        // unfortunately all H.323 based apps have a 16 bit length here, so
        // we allow for said longer length. There is presumably an addendum
        // to Q931 which describes this, and provides a means to discriminate
        // between the old 1 byte and the new 2 byte systems. However, at
        // present we assume it is always 2 bytes until we find something that
        // breaks it.
        len <<= 8;
        len |= data[offset++];

        // we also have a protocol discriminator, which we ignore
        offset++;
        len--;
      }

      if (offset + len > data.GetSize())
        return FALSE;

      memcpy(item->GetPointer(len), (const BYTE *)data+offset, len);
      offset += len;
    }

    fields.SetAt(discriminator, item);
  }

  return TRUE;
}


BOOL Q931::Encode(PBYTEArray & data) const
{
  PINDEX totalBytes = 5;
  unsigned discriminator;
  for (discriminator = 0; discriminator < 256; discriminator++) {
    if (fields.Contains(discriminator)) {
      if (discriminator < 128)
        totalBytes += fields[discriminator].GetSize() +
                            (discriminator != UserUserField ? 2 : 4);
      else
        totalBytes++;
    }
  }

  if (!data.SetMinSize(totalBytes))
    return FALSE;

  // Put in Q931 header
  PAssert(protocolDiscriminator < 256, PInvalidParameter);
  data[0] = (BYTE)protocolDiscriminator;
  data[1] = 2; // Length of call reference
  data[2] = (BYTE)(callReference >> 8);
  if (fromDestination)
    data[2] |= 0x80;
  data[3] = (BYTE)callReference;
  PAssert(messageType < 256, PInvalidParameter);
  data[4] = (BYTE)messageType;

  // The following assures disciminators are in ascending value order
  // as required by Q931 specification
  PINDEX offset = 5;
  for (discriminator = 0; discriminator < 256; discriminator++) {
    if (fields.Contains(discriminator)) {
      if (discriminator < 128) {
        int len = fields[discriminator].GetSize();

        if (discriminator != UserUserField) {
          data[offset++] = (BYTE)discriminator;
          data[offset++] = (BYTE)len;
        }
        else {
          len++; // Allow for protocol discriminator
          data[offset++] = (BYTE)discriminator;
          data[offset++] = (BYTE)(len >> 8);
          data[offset++] = (BYTE)len;
          len--; // Then put the length back again
          // We shall assume that the user-user field is an ITU protocol block (5)
          data[offset++] = 5;
        }

        memcpy(&data[offset], (const BYTE *)fields[discriminator], len);
        offset += len;
      }
      else
        data[offset++] = (BYTE)discriminator;
    }
  }

  return data.SetSize(offset);
}


static PString Q931FieldNameString(int number)
{
  switch (number) {
    case Q931::BearerCapabilityField :
      return "Bearer-Capability";
    case Q931::DisplayField :
      return "Display";
    case Q931::CalledPartyNumberField :
      return "Called-Party-Number";
    case Q931::UserUserField :
      return "User-User";
  }
  return psprintf("0x%02x", number);
}


void Q931::PrintOn(ostream & strm) const
{
  int indent = strm.precision() + 2;
  strm << "{\n"
       << setw(indent+24) << "protocolDiscriminator = " << protocolDiscriminator << '\n'
       << setw(indent+16) << "callReference = " << callReference << '\n'
       << setw(indent+7)  << "from = " << (fromDestination ? "destination" : "originator") << '\n'
       << setw(indent+14) << "messageType = " << GetMessageTypeName() << '\n';

  for (PINDEX i = 0; i < fields.GetSize(); i++)
    strm << setw(indent+7)
         << "field: " << Q931FieldNameString(fields.GetKeyAt(i)) << " = {\n"
         << hex << setfill('0')
         << setprecision(indent+2) << fields.GetDataAt(i) << '\n'
         << dec << setfill(' ')
         << setw(indent+2) << "}\n";

  strm << setw(indent-1) << "}";
}


PString Q931::GetMessageTypeName() const
{
  switch (messageType) {
    case AlertingMsg :
      return "Alerting";
    case CallProceedingMsg :
      return "CallProceeding";
    case ConnectMsg :
      return "Connect";
    case ConnectAckMsg :
      return "ConnectAck";
    case ProgressMsg :
      return "Progress";
    case SetupMsg :
      return "Setup";
    case SetupAckMsg :
      return "SetupAck";
    case ReleaseCompleteMsg :
      return "ReleaseComplete";
    case StatusMsg :
      return "Status";
    case NationalEscapeMsg :
      return "Escape";
    default :
      break;
  }

  return psprintf("<%u>", messageType);
}


void Q931::GenerateCallReference()
{
  static const char LastCallReference[] = "Last Call Reference";

  PConfig cfg("Globals");
  callReference = cfg.GetInteger(LastCallReference);

  if (callReference == 0)
    callReference = rand();
  else
    callReference++;

  callReference &= 0x7fff;

  if (callReference == 0)
    callReference = 1;

  cfg.SetInteger(LastCallReference, callReference);
}


BOOL Q931::HasField(FieldCodes field) const
{
  return fields.Contains(POrdinalKey(field));
}


PBYTEArray Q931::GetField(FieldCodes field) const
{
  if (fields.Contains(POrdinalKey(field)))
    return fields[field];

  return PBYTEArray();
}


void Q931::SetField(FieldCodes field, const PBYTEArray & userData)
{
  fields.SetAt(field, new PBYTEArray(userData));
}


void Q931::SetDisplayName(const PString & name)
{
  PBYTEArray bytes((const BYTE *)(const char *)name, name.GetLength()+1);
  SetField(DisplayField, bytes);
}


PString Q931::GetDisplayName() const
{
  if (!HasField(Q931::DisplayField))
    return PString();

  PBYTEArray display = GetField(Q931::DisplayField);
  if (display.IsEmpty())
    return PString();

  return PString((const char *)(const BYTE *)display, display.GetSize());
}


void Q931::SetCalledPartyNumber(const PString & number, unsigned plan, unsigned type)
{
  PINDEX len = number.GetLength();
  PBYTEArray bytes(len+1);
  bytes[0] = (BYTE)(0x80|((type&7)<<4)|(plan&15));
  memcpy(bytes.GetPointer()+1, (const char *)number, len);
  SetField(CalledPartyNumberField, bytes);
}


/////////////////////////////////////////////////////////////////////////////
