/*
 * transports.cxx
 *
 * H.323 transports handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: transports.cxx,v $
 * Revision 1.15  1999/09/02 15:25:39  robertj
 * Old GNU C ompiler compatibility
 *
 * Revision 1.14  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.13  1999/08/31 11:37:30  robertj
 * Fixed problem with apparently randomly losing signalling channel.
 *
 * Revision 1.12  1999/08/25 05:12:23  robertj
 * Changed MakeCall, so immediately spawns thread, no black on TCP connect.
 *
 * Revision 1.11  1999/07/23 02:37:53  robertj
 * Fixed problems with hang ups and crash closes of connections.
 *
 * Revision 1.10  1999/07/22 14:34:16  robertj
 * Fixed shut down problem, terminate packets being flushed on exit.
 *
 * Revision 1.9  1999/07/15 14:45:36  robertj
 * Added propagation of codec open error to shut down logical channel.
 * Fixed control channel start up bug introduced with tunnelling.
 *
 * Revision 1.8  1999/07/14 06:06:14  robertj
 * Fixed termination problems (race conditions) with deleting connection object.
 *
 * Revision 1.7  1999/07/09 06:09:52  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.6  1999/06/25 10:25:35  robertj
 * Added maintentance of callIdentifier variable in H.225 channel.
 *
 * Revision 1.5  1999/06/22 13:42:05  robertj
 * Added user question on listener version to accept incoming calls.
 *
 * Revision 1.4  1999/06/14 08:42:30  robertj
 * GNU C++ compatibility
 *
 * Revision 1.3  1999/06/13 12:41:14  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.2  1999/06/09 06:18:01  robertj
 * GCC compatibiltiy.
 *
 * Revision 1.1  1999/06/09 05:26:20  robertj
 * Major restructuring of classes.
 *
 */

#include <ptlib.h>
#include "h323.h"
#include "h323pdu.h"


class H225TransportThread : public PThread
{
  PCLASSINFO(H225TransportThread, PThread)

  public:
    H225TransportThread(H323EndPoint & endpoint,
                        H323Transport & transport);

  protected:
    void Main();

    H323Transport  & transport;
};


class H245TransportThread : public PThread
{
  PCLASSINFO(H245TransportThread, PThread)

  public:
    H245TransportThread(H323EndPoint & endpoint,
                        H323Connection & connection,
                        H323Transport & transport);

  protected:
    void Main();

    H323Connection & connection;
    H323Transport  & transport;
};


#define new PNEW


/////////////////////////////////////////////////////////////////////////////

H225TransportThread::H225TransportThread(H323EndPoint & endpoint,
                                         H323Transport & t)
  : PThread(endpoint.GetSignallingThreadStackSize(), NoAutoDeleteThread),
    transport(t)
{
  transport.AttachThread(this);
  Resume();
}


void H225TransportThread::Main()
{
  PTRACE(3, "H225\tStarted accept thread");

  H323Connection * connection = transport.HandleFirstSignallingChannelPDU();
  if (connection != NULL)
    connection->HandleSignallingChannel();
}


/////////////////////////////////////////////////////////////////////////////

H245TransportThread::H245TransportThread(H323EndPoint & endpoint,
                                         H323Connection & c,
                                         H323Transport & t)
  : PThread(endpoint.GetSignallingThreadStackSize(), NoAutoDeleteThread),
    connection(c),
    transport(t)
{
  transport.AttachThread(this);
  Resume();
}


void H245TransportThread::Main()
{
  PTRACE(3, "H245\tStarted thread");

  if (transport.AcceptControlChannel(connection))
    connection.HandleControlChannel();
}


/////////////////////////////////////////////////////////////////////////////

H323TransportAddress::H323TransportAddress(const char * cstr)
  : PString(cstr)
{
  Validate();
}


H323TransportAddress::H323TransportAddress(const PString & str)
  : PString(str)
{
  Validate();
}


H323TransportAddress::H323TransportAddress(const H225_TransportAddress & transport)
{
  switch (transport.GetTag()) {
    case H225_TransportAddress::e_ipAddress :
    {
      const H225_TransportAddress_ipAddress & ip = transport;
      sprintf("ip:%u.%u.%u.%u:%u",
              ip.m_ip[0], ip.m_ip[1], ip.m_ip[2], ip.m_ip[3],
              (unsigned)ip.m_port);
      break;
    }
  }
}


H323TransportAddress::H323TransportAddress(const H245_TransportAddress & transport)
{
  switch (transport.GetTag()) {
    case H245_TransportAddress::e_unicastAddress :
    {
      const H245_UnicastAddress & unicast = transport;
      switch (transport.GetTag()) {
        case H245_UnicastAddress::e_iPAddress :
        {
          const H245_UnicastAddress_iPAddress & ip = unicast;
          sprintf("ip:%u.%u.%u.%u:%u",
                  ip.m_network[0], ip.m_network[1], ip.m_network[2], ip.m_network[3],
                  (unsigned)ip.m_tsapIdentifier);
          break;
        }
      }
      break;
    }
  }
}


H323TransportAddress::H323TransportAddress(const PIPSocket::Address & ip, WORD port)
  : PString("ip:")
{
  *this += PIPSocket::GetHostName(ip);
  if (port != 0)
    sprintf(":%u", port);
}


void H323TransportAddress::Validate()
{
  if (IsEmpty())
    return;

  if (strncmp(theArray, "ip:", 3) == 0) {
    return;
  }

  *this = PString();
}


BOOL H323TransportAddress::SetPDU(H225_TransportAddress & pdu) const
{
  PIPSocket::Address ip;
  WORD port;
  if (GetIpAndPort(ip, port)) {
    pdu.SetTag(H225_TransportAddress::e_ipAddress);
    H225_TransportAddress_ipAddress & addr = pdu;
    for (PINDEX i = 0; i < 4; i++)
      addr.m_ip[i] = ip[i];
    addr.m_port = port;
    return TRUE;
  }


  return FALSE;
}


BOOL H323TransportAddress::SetPDU(H245_TransportAddress & pdu) const
{
  PIPSocket::Address ip;
  WORD port;
  if (GetIpAndPort(ip, port)) {
    pdu.SetTag(H245_TransportAddress::e_unicastAddress);

    H245_UnicastAddress & unicast = pdu;
    unicast.SetTag(H245_UnicastAddress::e_iPAddress);

    H245_UnicastAddress_iPAddress & addr = unicast;
    for (PINDEX i = 0; i < 4; i++)
      addr.m_network[i] = ip[i];
    addr.m_tsapIdentifier = port;
    return TRUE;
  }

  return FALSE;
}


BOOL H323TransportAddress::GetIpAndPort(PIPSocket::Address & ip,
                                        WORD & port,
                                        const char * proto) const
{
  if (strncmp(theArray, "ip:", 3) != 0) {
    PTRACE(2, "H323\tUse of non IP transport address: \"" << *this << '"');
    return FALSE;
  }

  PString host;
  PINDEX colon = Find(':', 3);
  if (colon == P_MAX_INDEX)
    host = Mid(3);
  else {
    host = Mid(3, colon-3);
    port = PIPSocket::GetPortByService(proto, Mid(colon+1));
  }

  if (host.IsEmpty() || port == 0) {
    PTRACE(2, "H323\tIllegal IP transport address: \"" << *this << '"');
    return FALSE;
  }

  if (PIPSocket::GetHostAddress(host, ip))
    return TRUE;

  PTRACE(1, "H323\tCould not find host : \"" << host << '"');
  return FALSE;
}


/////////////////////////////////////////////////////////////////////////////

H323Listener::H323Listener(H323EndPoint & end)
  : PThread(end.GetListenerThreadStackSize(), NoAutoDeleteThread),
    endpoint(end)
{
}


/////////////////////////////////////////////////////////////////////////////

H323Transport::H323Transport(H323EndPoint & end)
  : endpoint(end)
{
  thread = NULL;
}


H323Transport::~H323Transport()
{
  // Close the channel, but do not use Close() as that sets the indirect
  // channel members to NULL which crashes the other thread
  if (IsOpen())
    GetReadChannel()->Close();

  if (thread != NULL) {
    thread->WaitForTermination();
    delete thread;
  }
}


H323Connection * H323Transport::HandleFirstSignallingChannelPDU()
{
  PTRACE(3, "H225\tAwaiting first PDU");
  H323SignalPDU pdu;
  if (!pdu.Read(*this))
    return NULL;

  unsigned callReference = pdu.GetQ931().GetCallReference();
  PTRACE(3, "H225\tFirst PDU: callReference=" << callReference);

  H323Connection & connection = endpoint.GetConnection(*this, callReference, TRUE);

  connection.AttachSignalChannel(this);

  if (connection.HandleSignalPDU(pdu))
    return &connection;

  PTRACE(1, "H225\tSignal channel stopped on first PDU.");

  CloseSignallingChannel(callReference, connection.GetCallIdentifier());

  return NULL;
}


void H323Transport::CloseSignallingChannel(unsigned callReference,
                                           const H323GloballyUniqueID & callIdentifier)
{
  if (!IsOpen())
    return;

  PTRACE(2, "H225\tSending release complete PDU: callRef=" << callReference);
  H323SignalPDU replyPDU;
  replyPDU.BuildReleaseComplete(callReference, TRUE, callIdentifier);
  replyPDU.Write(*this);

  // Do not use Close() as this deletes the sub-channel member field crahsing
  // the background thread. Just close the sub-channel so breaks the threads
  // I/O block.
  GetReadChannel()->Close();
}


BOOL H323Transport::ConnectControlChannel(H323Connection & connection)
{
  if (!Connect())
    return FALSE;

  new H245TransportThread(endpoint, connection, *this);
  return TRUE;
}


void H323Transport::CloseControlChannel()
{
  // Do not use Close() as this deletes the sub-channel member field crahsing
  // the background thread. Just close the sub-channel so breaks the threads
  // I/O block.
  if (IsOpen())
    GetReadChannel()->Close();
}


void H323Transport::AttachThread(PThread * thrd)
{
  PAssert(thread == NULL, PLogicError);
  thread = thrd;
}


BOOL H323Transport::IsCompatibleTransport(const H225_TransportAddress & /*pdu*/) const
{
  PAssertAlways(PUnimplementedFunction);
  return FALSE;
}


void H323Transport::SetUpTransportPDU(H225_TransportAddress & /*pdu*/,
                                      BOOL /*localTsap*/) const
{
  PAssertAlways(PUnimplementedFunction);
}


void H323Transport::SetUpTransportPDU(H245_TransportAddress & /*pdu*/,
                                      unsigned /*port*/) const
{
  PAssertAlways(PUnimplementedFunction);
}


H323Transport * H323Transport::CreateControlChannel(H323Connection & /*connection*/)
{
  PAssertAlways(PUnimplementedFunction);
  return NULL;
}


BOOL H323Transport::AcceptControlChannel(H323Connection & /*connection*/)
{
  PAssertAlways(PUnimplementedFunction);
  return FALSE;
}


BOOL H323Transport::DiscoverGatekeepers(H323Gatekeeper & /*gk*/,
                                        const H323RasPDU & /*pdu*/)
{
  PAssertAlways(PUnimplementedFunction);
  return FALSE;
}


/////////////////////////////////////////////////////////////////////////////

H323ListenerTCP::H323ListenerTCP(H323EndPoint & end, WORD port)
  : H323Listener(end),
    listener(port)
{
}


H323ListenerTCP::~H323ListenerTCP()
{
  Close();
  WaitForTermination();
}


BOOL H323ListenerTCP::Open()
{
#ifdef _DEBUG
  return listener.Listen(5, 0, PSocket::CanReuseAddress);
#else
  return listener.Listen();
#endif
}


BOOL H323ListenerTCP::Close()
{
  return listener.Close();
}


void H323ListenerTCP::SetUpTransportPDU(H225_ArrayOf_TransportAddress & pdu) const
{
  PStringArray addresses = PIPSocket::GetHostAliases(PIPSocket::GetHostName());

  for (PINDEX i = 0; i < addresses.GetSize(); i++) {
    PIPSocket::Address ipAddr = addresses[i];
    if (ipAddr != 0 && ipAddr != PIPSocket::Address()) {
      PINDEX lastPos = pdu.GetSize();
      pdu.SetSize(lastPos+1);
      H225_TransportAddress & taddr = pdu[lastPos];

      taddr.SetTag(H225_TransportAddress::e_ipAddress);
      H225_TransportAddress_ipAddress & pduAddr = taddr;

      for (PINDEX j = 0; j < 4; j++)
        pduAddr.m_ip[j] = ipAddr[j];
      pduAddr.m_port = listener.GetPort();
    }
  }
}


void H323ListenerTCP::Main()
{
  PTRACE(2, "H323\tAwaiting TCP connections on port " << listener.GetPort());

  while (listener.IsOpen()) {
    PTCPSocket * signallingChannel = new PTCPSocket;
    if (signallingChannel->Accept(listener)) {
      H323TransportTCP * transport = new H323TransportTCP(endpoint);
      if (transport->Open(signallingChannel))
        new H225TransportThread(endpoint, *transport);
      else
        delete transport;
    }
    else {
      if (signallingChannel->GetErrorCode() != PChannel::Interrupted) {
        PError << "Q931\t\tAccept error:" << signallingChannel->GetErrorText();
        listener.Close();
      }
      delete signallingChannel;
    }
  }
}


/////////////////////////////////////////////////////////////////////////////

H323TransportIP::H323TransportIP(H323EndPoint & end, WORD remPort)
  : H323Transport(end),
    remoteAddress(0)
{
  PIPSocket::GetHostAddress(localAddress);
  localPort = 0;
  remotePort = remPort;
}


H323TransportAddress H323TransportIP::GetLocalAddress() const
{
  return H323TransportAddress(localAddress, localPort);
}


H323TransportAddress H323TransportIP::GetRemoteAddress() const
{
  return H323TransportAddress(remoteAddress, remotePort);
}


BOOL H323TransportIP::IsCompatibleTransport(const H225_TransportAddress & pdu) const
{
  return pdu.GetTag() == H225_TransportAddress::e_ipAddress;
}


void H323TransportIP::SetUpTransportPDU(H225_TransportAddress & pdu, BOOL localTsap) const
{
  pdu.SetTag(H225_TransportAddress::e_ipAddress);
  H225_TransportAddress_ipAddress & pduAddr = pdu;

  PIPSocket::Address ipAddr = localTsap ? localAddress : remoteAddress;
  for (PINDEX i = 0; i < 4; i++)
    pduAddr.m_ip[i] = ipAddr[i];
  pduAddr.m_port = localTsap ? localPort : remotePort;
}


void H323TransportIP::SetUpTransportPDU(H245_TransportAddress & pdu, unsigned port) const
{
  pdu.SetTag(H245_TransportAddress::e_unicastAddress);

  H245_UnicastAddress & unicast = pdu;
  unicast.SetTag(H245_UnicastAddress::e_iPAddress);

  H245_UnicastAddress_iPAddress & addr = unicast;
  for (PINDEX i = 0; i < 4; i++)
    addr.m_network[i] = localAddress[i];
  addr.m_tsapIdentifier = port;
}


void H323TransportIP::FindLocalAddress()
{
  PIPSocket::RouteTable table;
  if (PIPSocket::GetRouteTable(table)) {
    PINDEX routeIdx = table.GetSize();
    while (routeIdx-- > 0) {
      PIPSocket::RouteEntry route = table[routeIdx];
      if ((remoteAddress&route.GetNetMask()) == route.GetNetwork()) {
        localAddress = route.GetDestination();
        break;
      }
    }
  }
}


/////////////////////////////////////////////////////////////////////////////

H323TransportTCP::H323TransportTCP(H323EndPoint & end)
  : H323TransportIP(end, H323ListenerTCP::DefaultSignalPort)
{
  h245listener = NULL;
}


H323TransportTCP::~H323TransportTCP()
{
  delete h245listener;  // Delete and H245 listener that may be present
}


BOOL H323TransportTCP::OnOpen()
{
  PIPSocket * socket = (PIPSocket *)GetReadChannel();

  // Get name of the remote computer for information purposes
  if (!socket->GetPeerAddress(remoteAddress, remotePort))
    return FALSE;

  // get local address of incoming socket to ensure that multi-homed machines
  // use a NIC address that is guaranteed to be addressable to destination
  if (!socket->GetLocalAddress(localAddress, localPort))
    return FALSE;

  // make sure do not lose outgoing packets on close
  const linger ling = { 1, 10 };
  return socket->SetOption(SO_LINGER, &ling, sizeof(ling));
}


BOOL H323TransportTCP::SetRemoteAddress(const H323TransportAddress & address)
{
  if (!address.GetIpAndPort(remoteAddress, remotePort, "tcp"))
    return FALSE;

  FindLocalAddress();
  return TRUE;
}


BOOL H323TransportTCP::ReadPDU(PBYTEArray & pdu)
{
  // Get RFC1006 TPKT length
  BYTE tpkt[4];
  if (ReadBlock(tpkt, sizeof(tpkt)) && tpkt[0] == 3) { // Only support version 3
    int packetLength = ((tpkt[2] << 8)|tpkt[3]) - 4;
    if (ReadBlock(pdu.GetPointer(packetLength), packetLength))
      return TRUE;
  }

  pdu.SetSize(0);
  return FALSE;
}


BOOL H323TransportTCP::WritePDU(const PBYTEArray & pdu)
{
  int packetLength = pdu.GetSize() + 4;

  // Send RFC1006 TPKT length
  BYTE tpkt[4];
  tpkt[0] = 3;
  tpkt[1] = 0;
  tpkt[2] = (BYTE)(packetLength >> 8);
  tpkt[3] = (BYTE)packetLength;

  return Write(tpkt, sizeof(tpkt)) && Write((const BYTE *)pdu, pdu.GetSize());
}


BOOL H323TransportTCP::Connect()
{
  PTCPSocket * socket = new PTCPSocket(remotePort);
  if (!socket->Connect(remoteAddress)) {
    PTRACE(1, "TCP\tCould not connect to "
           << remoteAddress << ':' << remotePort << ' ' << socket->GetErrorText());
    delete socket;
    return FALSE;
  }

  if (!Open(socket))
    return FALSE;

  PTRACE(1, "TCP\tStarted connection to "
         << remoteAddress << ':' << remotePort
         << " (if=" << localAddress << ':' << localPort << ')');

  return TRUE;
}


H323Transport * H323TransportTCP::CreateControlChannel(H323Connection & connection)
{
  // construct H245 socket
  PTCPSocket * socket = new PTCPSocket;
  if (!socket->Listen()) {
    PTRACE(1, "H225\tTCP Listen for H245 failed: " << socket->GetErrorText());
    delete socket;
    return NULL;
  }

  PTRACE(3, "H225\tTCP Listen for H245 on "
                   << localAddress << ':' << socket->GetPort());

  // Make H323 Connect PDU
  H323SignalPDU connectPDU;
  connectPDU.BuildConnect(endpoint,
                          connection.GetCallReference(),
                          connection.GetCallIdentifier(),
                          connection.GetConferenceIdentifier(),
                          localAddress, socket->GetPort());

  // Send H323 Connect PDU
  if (!connection.WriteSignalPDU(connectPDU)) {
    delete socket;
    return NULL;
  }

  H323TransportTCP * transport = new H323TransportTCP(endpoint);
  transport->h245listener = socket;
  new H245TransportThread(endpoint, connection, *transport);
  return transport;
}


BOOL H323TransportTCP::AcceptControlChannel(H323Connection & connection)
{
  if (IsOpen())
    return TRUE;

  if (h245listener == NULL) {
    PAssertAlways(PLogicError);
    return FALSE;
  }

  PTRACE(3, "H245\tTCP Accept wait");

  PTCPSocket * h245Socket = new PTCPSocket;

  h245listener->SetReadTimeout(30000); // Wait 30 seconds for remote H245 connect
  if (h245Socket->Accept(*h245listener))
    return Open(h245Socket);

  PTRACE(1, "H225\tAccept for H245 failed: " << h245Socket->GetErrorText());
  delete h245Socket;

  H323SignalPDU replyPDU;
  replyPDU.BuildReleaseComplete(connection.GetCallReference(), TRUE,
                                connection.GetCallIdentifier());
  connection.WriteSignalPDU(replyPDU);
  return FALSE;
}


/////////////////////////////////////////////////////////////////////////////

H323TransportUDP::H323TransportUDP(H323EndPoint & end)
  : H323TransportIP(end, DefaultRasPort)
{
  PUDPSocket * udp = new PUDPSocket;
  udp->Listen();
  Open(udp);

  localPort = udp->GetPort();
}


BOOL H323TransportUDP::SetRemoteAddress(const H323TransportAddress & address)
{
  if (!address.GetIpAndPort(remoteAddress, remotePort, "udp"))
    return FALSE;

  FindLocalAddress();
  return TRUE;
}


BOOL H323TransportUDP::Connect()
{
  if (remoteAddress == 0 || remotePort == 0)
    return FALSE;

  PUDPSocket * socket = (PUDPSocket *)GetReadChannel();
  socket->SetSendAddress(remoteAddress, remotePort);

  return TRUE;
}


BOOL H323TransportUDP::ReadPDU(PBYTEArray & pdu)
{
  if (!Read(pdu.GetPointer(10000), 10000)) {
    pdu.SetSize(0);
    return FALSE;
  }

  pdu.SetSize(GetLastReadCount());
  return TRUE;
}


BOOL H323TransportUDP::WritePDU(const PBYTEArray & pdu)
{
  return Write((const BYTE *)pdu, pdu.GetSize());
}


BOOL H323TransportUDP::DiscoverGatekeepers(H323Gatekeeper & gk,
                                           const H323RasPDU & request)
{
  PTRACE(3, "H225\tGatekeeper discovery started");

  if (!IsOpen())
    return FALSE;

  remoteAddress = 0;
  remotePort = 0;

  PUDPSocket * socket = (PUDPSocket *)GetReadChannel();
  if (!socket->SetOption(SO_BROADCAST, 1))
    return FALSE;

  socket->SetSendAddress(INADDR_BROADCAST, DefaultRasPort);

  if (!request.Write(*this))
    return FALSE;

  socket->SetOption(SO_BROADCAST, 0);

  H323RasPDU response;
  if (!response.Read(*this)) {
    PTRACE(3, "RAS\tError on discover request read");
    return FALSE;
  }

  do {
    if (gk.HandleRasPDU(response)) {
      socket->GetLastReceiveAddress(remoteAddress, remotePort);
      socket->SetSendAddress(remoteAddress, remotePort);
      FindLocalAddress();
      return TRUE;
    }
  } while (response.Read(*this));

  return FALSE;
}


/////////////////////////////////////////////////////////////////////////////
