/*
 * xjack.cxx
 *
 * Application source for QuickNet Internet PhoneJack H.323 client
 *
 * Copyright (c) 1993-1999 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: xjack.cxx,v $
 * Revision 1.3  1999/08/27 02:12:25  robertj
 * Added missing AwaitingTermiantion override.
 *
 * Revision 1.2  1999/08/25 05:17:44  robertj
 * Simplified application structure (better for tutorial purposes)
 *
 * Revision 1.1  1999/07/26 07:06:53  craigs
 * Initial version with support for Quicknet cards
 *
 */

#include <ptlib.h>
#include "main.h"
#include "version.h"
#include "xjack.h"

#include <h245.h>

#ifndef _WIN32
#include <ixjuser.h>
#endif


#define SAVE_RTP


/*
  This table defines the parameters for the codecs available on the xJack cards.

  Entries in this table must correspond to classes descended from XJackCodec as defined
  in main.h. The indexes into this table must correspond to the enum defined in XJackCodec

*/

enum {
  CodecInfo_Ioctl,
  CodecInfo_BlockSize,
  CodecInfo_SamplesPerBlock,
  CodecInfo_Depth,
  CodecInfo_RTPPayloadType,
  CodecInfo_CapabilitySubtype,
  CodecInfo_EntrySize
};

static int codecInfo[][CodecInfo_EntrySize] = {
//   QuickNet     read/write   samples   codec       RTP                H.323
//    ioctl         block        per     buffer     payload            capability
//    code          size        block    depth       type               subtype

  {   ULAW,         240,         240,    2,     RTP_DataFrame::PCMU,   H245_AudioCapability::e_g711Ulaw64k },
  {   G723_63,       24,         240,    10,    RTP_DataFrame::G7231,  H245_AudioCapability::e_g7231 }
};

static char * codecDescriptions[CodecInfo_EntrySize] = {
  "G711 uLaw",
  "G723.1 6.3k"
};

#define MAX_CODEC_INDEX (sizeof(codecInfo)/sizeof(codecInfo[0]))


BOOL XJackH323EndPoint::InitialiseCodecs(PArgList & args)
{
  int deviceNumber = args.GetOptionString('q').AsInteger();

  deviceName = psprintf("/dev/ixj%i", deviceNumber);
  openCount = 0;

  if ((ixj_handle = ::open((const char *)deviceName, O_RDWR)) < 0) {
    cerr << "Cannot open device \"" << deviceName << "\"" << endl;
    return FALSE;
  }

  cout << "xJack card: " << deviceName << "\n";

  H323Capability * g711ulaw64k = new XJackG711uLaw64kCapability;
  //H323Capability * g7231       = new XJackG7231Capability;

  if (args.HasOption("g711-ulaw"))
    AddCapability(g711ulaw64k);
  //if (args.HasOption("g7231-63k"))
  //  AddCapability(g7231);

  //AddCapability(g7231);
  AddCapability(g711ulaw64k);

  //SetCapability(0, 0, g7231);
  SetCapability(0, 0, g711ulaw64k);

  return TRUE;
}


H323Connection * XJackH323EndPoint::CreateConnection(unsigned callReference)
{
  return new XJackH323Connection(*this, callReference, noFastStart);
}


void XJackH323EndPoint::MakeOutgoingCall(const PString & dest, const PString & proxy)
{
  // wait for the handset to go offhook
  while (IsOnHook()) {
    cout << "Lift handset" << endl;
    PThread::Current()->Sleep(1000);
  }

  H323TransportTCP * transport = new H323TransportTCP(*this);
  if (!proxy)
    transport->SetGatewayName(proxy);

  cout << GetLocalUserName() << " is calling " << dest << endl;

  PString token;
  if (!MakeCall(dest, transport, token)) {
    cerr << "Remote H.323 party \"" << dest << "\" refused the call or did not respond.\n";
    return;
  }

  remoteUsername = dest;
}


void XJackH323EndPoint::ListenForIncomingCalls()
{
  // wait for the handset to be onhook
  while (!IsOnHook()) {
    cout << "Put handset on hook" << endl;
    PThread::Current()->Sleep(1000);
  }

  H323ListenerTCP * listener = new H323ListenerTCP(*this);
  if (!StartListener(listener)) {
    cerr <<  "Could not open H.323 listener port on " << listener->GetListenerPort() << endl;
    return;
  }

  cout << "Waiting for incoming calls for \"" << GetLocalUserName() << "\"\n";
  PThread::Current()->Sleep(PMaxTimeInterval);
}


void XJackH323EndPoint::AwaitTermination()
{
  // wait for handset to go onhook
  while (!IsOnHook()) {
    cerr << "Checking for handset to go onhook" << endl;
    PThread::Current()->Sleep(1000);
  }

  MyH323EndPoint::AwaitTermination();
}


BOOL XJackH323EndPoint::RingHandset(int num)
{
  // set the maximum number of rings to 1
  ::ioctl(ixj_handle, IXJCTL_MAXRINGS, 1);

  while (num > 0) {

    // ring once, and see if answered
    if (::ioctl(ixj_handle, IXJCTL_RING) != 0) 
      return TRUE;
  }

  return FALSE;
}

BOOL XJackH323EndPoint::IsOnHook()
{
  return ::ioctl(ixj_handle, IXJCTL_HOOKSTATE) == 0;
}


int XJackH323EndPoint::Read(BYTE * buffer, unsigned len)
{
  readBuffer = buffer;
  readLen    = len;

  mutex.Wait();

  while (readBuffer != NULL) 
    ReadWrite();

  mutex.Signal();

  return readStat;
}

int XJackH323EndPoint::Write(const BYTE * buffer, unsigned len)
{
  writeBuffer = buffer;
  writeLen    = len;

  mutex.Wait();

  while (writeBuffer != NULL) 
    ReadWrite();

  mutex.Signal();

  return writeStat;
}


void XJackH323EndPoint::ReadWrite()
{
  fd_set(rfds);
  fd_set(wfds);
  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  FD_SET(ixj_handle, &rfds);
  FD_SET(ixj_handle, &wfds);
  ::select(ixj_handle+1, &rfds, &wfds, NULL, NULL);

  if (readBuffer != NULL && FD_ISSET(ixj_handle, &rfds)) {
    readStat = ::read(ixj_handle, readBuffer, readLen);
    readBuffer = NULL;
  }

  if (writeBuffer != NULL && FD_ISSET(ixj_handle, &wfds)) {
    writeStat = ::write(ixj_handle, writeBuffer, writeLen);
    writeBuffer = NULL;
  }
}

///////////////////////////////////////////////////////////////

XJackH323Connection::XJackH323Connection(XJackH323EndPoint & ep,
                                         unsigned callReference,
                                         BOOL _noFastStart)
  : MyH323Connection(ep, callReference, _noFastStart),
    xjEndpoint(ep)
{
}


BOOL XJackH323Connection::OnAnswerCall(const PString & caller,
                                    const H323SignalPDU & /*setupPDU*/,
                                    H323SignalPDU & /*connectPDU*/)
{
  PTRACE(1, "H225\tOnWaitForAnswer - user input");

  // ring the handset
  BOOL answered = xjEndpoint.RingHandset(10);

  if (!answered)
    cout << "Rejecting call." << endl;
  else
    cout << "Accepting call." << endl;

  xjEndpoint.remoteUsername = caller;

  return answered;
}


///////////////////////////////////////////////////////////////

XJackCodec::XJackCodec(Direction dir, PINDEX _codecIndex)
  : H323Codec(dir),
    handle(-1),
    codecIndex(_codecIndex)
{
  PAssert(codecIndex < MAX_CODEC_INDEX, "invalid codec specified");
  xjackEndp = NULL;

  sampleBlockSize = codecInfo[codecIndex][CodecInfo_BlockSize];
  sampleCount     = codecInfo[codecIndex][CodecInfo_SamplesPerBlock];
  rtpPayloadType  = codecInfo[codecIndex][CodecInfo_RTPPayloadType];
  PTRACE(3, "Creating XJackCodec");
}


BOOL XJackCodec::Open(H323Connection & connection)
{
  // get endpoint, as this contains the handle
  xjackEndp = &(XJackH323EndPoint &)connection.GetEndPoint();

#ifdef SAVE_RTP
  if (direction == Encoder) {
    rtpFile.Open("/tmp/rtp.in", PFile::WriteOnly);
  } else {
    rtpFile.Open("/tmp/rtp.out", PFile::WriteOnly);
  }
#endif

  // get the handle for the device
  handle = xjackEndp->GetHandle();
  if (xjackEndp->GetOpenCount(1) == 1) {

    PError << "XJack device " << xjackEndp->GetDeviceName() << " opened" << endl;

    // reset the DSP, just in case
    ::ioctl(handle, IXJCTL_DSP_RESET);

    // set all of the paramaters
    int codecIoctl = codecInfo[codecIndex][CodecInfo_Ioctl];
    int depth      = codecInfo[codecIndex][CodecInfo_Depth];

    cout << "Setting XJack codec to " << codecIoctl << endl;

    ::ioctl(handle, IXJCTL_REC_CODEC, codecIoctl);
    ::ioctl(handle, IXJCTL_PLAY_CODEC, codecIoctl);

    ::ioctl(handle, IXJCTL_REC_DEPTH, depth);
    ::ioctl(handle, IXJCTL_PLAY_DEPTH, depth);

    ::ioctl(handle, IXJCTL_REC_START);
    ::ioctl(handle, IXJCTL_PLAY_START);
    ::ioctl(handle, IXJCTL_AEC_START);
  }
  return TRUE;
}


XJackCodec::~XJackCodec()
{
  Close();
}

void XJackCodec::Close()
{
  if (xjackEndp != NULL)
    xjackEndp->GetOpenCount(-1);
}


BOOL XJackCodec::Read(BYTE * buffer, unsigned & length)
{
#ifdef DISABLE_READ
  return FALSE;
#else

  int stat = xjackEndp->Read(buffer, sampleBlockSize);

  if (stat != (signed)sampleBlockSize) {
    PTRACE(1, "XJack\tRead failed: " << stat);
    return FALSE;
  }

  if (rtpFile.IsOpen())
    rtpFile.Write(buffer, length);

  length = sampleBlockSize;

  return TRUE;
#endif
}


BOOL XJackCodec::Write(const BYTE * buffer, unsigned length, unsigned & writtenLength)
{
#ifdef DISABLE_WRITE
  return FALSE;
#else

  if (rtpFile.IsOpen())
    rtpFile.Write(buffer, length);

  writtenLength = sampleBlockSize;

  if (length == 0) 
    return TRUE;

  xjackEndp->Write(buffer, sampleBlockSize);

  return TRUE;
#endif
}

///////////////////////////////////////////////////////////////

XJackCapability::XJackCapability(PINDEX _codecIndex)
  : codecIndex(_codecIndex)
{
  subType = codecInfo[codecIndex][CodecInfo_CapabilitySubtype];
  PTRACE(3, "Creating XJackCapability with codecIndex " << codecIndex);
}

void XJackCapability::PrintOn(ostream & strm) const
{
  strm << "XJack : "
       << codecDescriptions[codecIndex];
  H323AudioCapability::PrintOn(strm);
}


BOOL XJackCapability::OnSendingPDU(H245_AudioCapability & cap) const
{
  // set the choice to the correct type
  cap.SetTag(subType);

  return EncodeSendingPDU(cap);
}

BOOL XJackG7231Capability::EncodeSendingPDU(H245_AudioCapability & cap) const
{
  // get choice data
  H245_AudioCapability_g7231 & g7231 = cap;

  // max number of audio frames per PDU (??)
  g7231.m_maxAl_sduAudioFrames = 12;

  // no silence suppression
  g7231.m_silenceSuppression   = FALSE;

  return TRUE;
}

BOOL XJackG7231Capability::OnReceivedPDU(const H245_AudioCapability & )
{
  return TRUE;
}


BOOL XJackG711uLaw64kCapability::EncodeSendingPDU(H245_AudioCapability & pdu) const
{
  PASN_Integer & value = pdu;

  // indicate that we want the the other end to send 30ms of data at a time
  value = 30;

  return TRUE;
}

BOOL XJackG711uLaw64kCapability::OnReceivedPDU(const H245_AudioCapability & )
{
  return TRUE;
}

///////////////////////////////////////////////////////////////

H323Codec * XJackCapability::CreateCodec(H323Codec::Direction direction) const
{
  return new XJackCodec(direction, codecIndex);
}


// End of File ///////////////////////////////////////////////////////////////
